"""Quaxed :external:`operator`.

This module wraps the functions in :external:`operator` with
:external:`quax.quaxify`. The wrapping happens dynamically through a
module-level ``__dir__`` and ``__getattr__``. The list of available functions is
in ``__all__`` and documented in the built-in :external:`operator` library.

"""

import operator
import sys
from collections.abc import Callable
from typing import Any

from quax import quaxify

__all__ = operator.__all__


def __dir__() -> list[str]:
    """List the module contents."""
    return sorted(__all__)


# TODO: return type hint signature
def __getattr__(name: str) -> Callable[..., Any]:
    """Get the :external:`quax.quaxify`'ed function."""
    # Quaxify the operator
    out = quaxify(getattr(operator, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out



"""Test with JAX inputs."""

import tempfile

import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jt
import numpy as np
import pytest

import quaxed.numpy as qnp


@pytest.fixture
def x1():
    """Test input."""
    return jnp.array([1, 2, 3], dtype=float)


@pytest.fixture
def x2():
    """Test input."""
    return jnp.array([4, 5, 6], dtype=float)


@pytest.fixture
def xbool():
    """Test input."""
    return jnp.array([True, False, True], dtype=bool)


###############################################################################


@pytest.mark.parametrize("name", qnp._core._DIRECT_TRANSFER)  # noqa: SLF001
def test_direct_transfer(name):
    """Test direct transfers."""
    assert getattr(qnp, name) is getattr(jnp, name)


###############################################################################


def test_abs():
    """Test `quaxed.numpy.abs`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.abs(x) == jnp.abs(x))


def test_absolute():
    """Test `quaxed.numpy.absolute`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.absolute(x) == jnp.absolute(x))


def test_acos():
    """Test `quaxed.numpy.acos`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.acos(x) == jnp.acos(x))


def test_acosh():
    """Test `quaxed.numpy.acosh`."""
    x = jnp.array([1, 2, 3], dtype=float)
    assert jnp.all(qnp.acosh(x) == jnp.acosh(x))


def test_add(x1, x2):
    """Test `quaxed.numpy.add`."""
    assert jnp.all(qnp.add(x1, x2) == jnp.add(x1, x2))


def test_all(x1):
    """Test `quaxed.numpy.all`."""
    x = jnp.array([True, False, True], dtype=bool)
    assert qnp.all(x) == jnp.all(x)


def test_allclose(x1, x2):
    """Test `quaxed.numpy.allclose`."""
    x = jnp.array([1, 2, 3], dtype=float)
    y = jnp.array([1, 2, 3], dtype=float) + 1e-9
    assert qnp.allclose(x, y) == jnp.allclose(x, y)

    # allclose
    assert qnp.allclose(x1, x2, atol=1e-8) == jnp.allclose(x1, x2, atol=1e-8)


def test_amax(x1):
    """Test `quaxed.numpy.amax`."""
    assert qnp.amax(x1) == jnp.amax(x1)


def test_amin(x1):
    """Test `quaxed.numpy.amin`."""
    assert qnp.amin(x1) == jnp.amin(x1)


def test_angle(x1):
    """Test `quaxed.numpy.angle`."""
    assert jnp.all(qnp.angle(x1) == jnp.angle(x1))


def test_array_equal(x1, x2):
    """Test `quaxed.numpy.array_equal`."""
    assert qnp.array_equal(x1, x2) == jnp.array_equal(x1, x2)


def test_any(x1):
    """Test `quaxed.numpy.any`."""
    x = jnp.array([True, False, True], dtype=bool)
    assert qnp.any(x) == jnp.any(x)


def test_append(x1):
    """Test `quaxed.numpy.append`."""
    assert jnp.all(qnp.append(x1, 4) == jnp.append(x1, 4))


def test_apply_along_axis(x1):
    """Test `quaxed.numpy.apply_along_axis`."""
    assert jnp.all(
        qnp.apply_along_axis(lambda x: x + 1, 0, x1)
        == jnp.apply_along_axis(lambda x: x + 1, 0, x1)
    )


def test_apply_over_axes(x1):
    """Test `quaxed.numpy.apply_over_axes`."""
    assert jnp.all(
        qnp.apply_over_axes(lambda x, _: x + 1, x1, [0, 1])
        == jnp.apply_over_axes(lambda x, _: x + 1, x1, [0, 1])
    )


def test_arange():
    """Test `quaxed.numpy.arange`."""
    assert jnp.all(qnp.arange(3) == jnp.arange(3))


def test_arccos():
    """Test `quaxed.numpy.arccos`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.arccos(x) == jnp.arccos(x))


def test_arccosh():
    """Test `quaxed.numpy.arccosh`."""
    x = jnp.array([1, 2, 3], dtype=float)
    assert jnp.all(qnp.arccosh(x) == jnp.arccosh(x))


def test_arcsin():
    """Test `quaxed.numpy.arcsin`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.arcsin(x) == jnp.arcsin(x))


def test_arcsinh():
    """Test `quaxed.numpy.arcsinh`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.arcsinh(x) == jnp.arcsinh(x))


def test_arctan():
    """Test `quaxed.numpy.arctan`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.arctan(x) == jnp.arctan(x))


def test_arctan2(x1, x2):
    """Test `quaxed.numpy.arctan2`."""
    assert jnp.all(qnp.arctan2(x1, x2) == jnp.arctan2(x1, x2))


def test_arctanh():
    """Test `quaxed.numpy.arctanh`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.arctanh(x) == jnp.arctanh(x))


def test_argmax(x1):
    """Test `quaxed.numpy.argmax`."""
    assert qnp.argmax(x1) == jnp.argmax(x1)


def test_argmin(x1):
    """Test `quaxed.numpy.argmin`."""
    assert qnp.argmin(x1) == jnp.argmin(x1)


def test_argpartition(x1):
    """Test `quaxed.numpy.argpartition`."""
    assert jnp.all(qnp.argpartition(x1, 1) == jnp.argpartition(x1, 1))


def test_argsort(x1):
    """Test `quaxed.numpy.argsort`."""
    assert jnp.all(qnp.argsort(x1) == jnp.argsort(x1))


def test_argwhere(x1):
    """Test `quaxed.numpy.argwhere`."""
    assert jnp.all(qnp.argwhere(x1) == jnp.argwhere(x1))


def test_around(x1):
    """Test `quaxed.numpy.around`."""
    assert jnp.all(qnp.around(x1) == jnp.around(x1))


def test_array(x1):
    """Test `quaxed.numpy.array`."""
    assert jnp.all(qnp.array(x1) == jnp.array(x1))


def test_array_equal(x1, x2):
    """Test `quaxed.numpy.array_equal`."""
    assert qnp.array_equal(x1, x2) == jnp.array_equal(x1, x2)


def test_array_equiv(x1, x2):
    """Test `quaxed.numpy.array_equiv`."""
    assert qnp.array_equiv(x1, x2) == jnp.array_equiv(x1, x2)


@pytest.mark.xfail(reason="Not implemented.")
def test_array_str(x1):
    """Test `quaxed.numpy.array_str`."""
    assert qnp.array_str(x1) == jnp.array_str(x1)


def test_asarray():
    """Test `quaxed.numpy.asarray`."""
    x = [1.0, 2, 3, 4]
    assert jnp.all(qnp.asarray(x) == jnp.asarray(x))


def test_asin():
    """Test `quaxed.numpy.asin`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.asin(x) == jnp.asin(x))


def test_asinh():
    """Test `quaxed.numpy.asinh`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.asinh(x) == jnp.asinh(x))


def test_astype(x1):
    """Test `quaxed.numpy.astype`."""
    assert jnp.all(qnp.astype(x1, int) == jnp.astype(x1, int))


def test_atan():
    """Test `quaxed.numpy.atan`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.arctan(x) == jnp.atan(x))


def test_atan2(x1, x2):
    """Test `quaxed.numpy.atan2`."""
    assert jnp.all(qnp.arctan2(x1, x2) == jnp.atan2(x1, x2))


def test_atanh():
    """Test `quaxed.numpy.atanh`."""
    x = jnp.array([-1, 0, 1], dtype=float)
    assert jnp.all(qnp.arctanh(x) == jnp.atanh(x))


def test_atleast_1d(x1):
    """Test `quaxed.numpy.atleast_1d`."""
    assert jnp.all(qnp.atleast_1d(x1) == jnp.atleast_1d(x1))


def test_atleast_2d(x1):
    """Test `quaxed.numpy.atleast_2d`."""
    assert jnp.all(qnp.atleast_2d(x1) == jnp.atleast_2d(x1))


def test_atleast_3d(x1):
    """Test `quaxed.numpy.atleast_3d`."""
    assert jnp.all(qnp.atleast_3d(x1) == jnp.atleast_3d(x1))


def test_average(x1):
    """Test `quaxed.numpy.average`."""
    assert qnp.average(x1) == jnp.average(x1)


def test_bartlett():
    """Test `quaxed.numpy.bartlett`."""
    assert jnp.all(qnp.bartlett(3) == jnp.bartlett(3))


def test_bincount():
    """Test `quaxed.numpy.bincount`."""
    x = jnp.asarray([0, 1, 1, 2, 2, 2])
    assert jnp.all(qnp.bincount(x) == jnp.bincount(x))


def test_bitwise_and(xbool):
    """Test `quaxed.numpy.bitwise_and`."""
    assert jnp.all(qnp.bitwise_and(xbool, xbool) == jnp.bitwise_and(xbool, xbool))


def test_bitwise_count(xbool):
    """Test `quaxed.numpy.bitwise_count`."""
    assert jnp.all(qnp.bitwise_count(xbool) == jnp.bitwise_count(xbool))


def test_bitwise_invert(xbool):
    """Test `quaxed.numpy.bitwise_invert`."""
    assert jnp.all(qnp.bitwise_invert(xbool) == jnp.bitwise_invert(xbool))


def test_bitwise_left_shift(xbool):
    """Test `quaxed.numpy.bitwise_left_shift`."""
    assert jnp.all(qnp.bitwise_left_shift(xbool, 1) == jnp.bitwise_left_shift(xbool, 1))


def test_bitwise_not(xbool):
    """Test `quaxed.numpy.bitwise_not`."""
    assert jnp.all(qnp.bitwise_not(xbool) == jnp.bitwise_not(xbool))


def test_bitwise_or(xbool):
    """Test `quaxed.numpy.bitwise_or`."""
    assert jnp.all(qnp.bitwise_or(xbool, xbool) == jnp.bitwise_or(xbool, xbool))


def test_bitwise_right_shift(xbool):
    """Test `quaxed.numpy.bitwise_right_shift`."""
    assert jnp.all(
        qnp.bitwise_right_shift(xbool, 1) == jnp.bitwise_right_shift(xbool, 1)
    )


def test_bitwise_xor(xbool):
    """Test `quaxed.numpy.bitwise_xor`."""
    assert jnp.all(qnp.bitwise_xor(xbool, xbool) == jnp.bitwise_xor(xbool, xbool))


def test_blackman():
    """Test `quaxed.numpy.blackman`."""
    assert jnp.all(qnp.blackman(3) == jnp.blackman(3))


def test_block(x1):
    """Test `quaxed.numpy.block`."""
    assert jnp.all(qnp.block([x1, x1]) == jnp.block([x1, x1]))


def test_bool():
    """Test `quaxed.numpy.bool`."""
    assert qnp.bool(1) == jnp.bool(1)


def test_bool_():
    """Test `quaxed.numpy.bool_`."""
    assert qnp.bool_(1) == jnp.bool_(1)


def test_broadcast_arrays(x1, x2):
    """Test `quaxed.numpy.broadcast_arrays`."""
    got1, got3 = qnp.broadcast_arrays(x1, x2)
    expected1, expected2 = jnp.broadcast_arrays(x1, x2)
    assert jnp.all(got1 == expected1)
    assert jnp.all(got3 == expected2)


def test_broadcast_shapes(x1, x2):
    """Test `quaxed.numpy.broadcast_shapes`."""
    assert jnp.all(
        qnp.broadcast_shapes(x1.shape, x2.shape)
        == jnp.broadcast_shapes(x1.shape, x2.shape)
    )


def test_broadcast_to(x1):
    """Test `quaxed.numpy.broadcast_to`."""
    assert jnp.all(qnp.broadcast_to(x1, (3, 3)) == jnp.broadcast_to(x1, (3, 3)))


@pytest.mark.xfail(reason="Not implemented.")
def test_c_():
    """Test `quaxed.numpy.c_`."""
    assert jnp.all(qnp.c_[1:3, 4:6] == jnp.c_[1:3, 4:6])


def test_can_cast(x1):
    """Test `quaxed.numpy.can_cast`."""
    assert qnp.can_cast(x1, int) == jnp.can_cast(x1, int)


def test_cbrt(x1):
    """Test `quaxed.numpy.cbrt`."""
    assert jnp.all(qnp.cbrt(x1) == jnp.cbrt(x1))


def test_ceil(x1):
    """Test `quaxed.numpy.ceil`."""
    assert jnp.all(qnp.ceil(x1) == jnp.ceil(x1))


def test_choose(x1):
    """Test `quaxed.numpy.choose`."""
    assert jnp.all(qnp.choose(0, [x1, x1]) == jnp.choose(0, [x1, x1]))


def test_clip(x1):
    """Test `quaxed.numpy.clip`."""
    assert jnp.all(qnp.clip(x1, 1, 2) == jnp.clip(x1, 1, 2))


def test_column_stack(x1):
    """Test `quaxed.numpy.column_stack`."""
    assert jnp.all(qnp.column_stack([x1, x1]) == jnp.column_stack([x1, x1]))


@pytest.mark.skip(reason="Not available.")
def test_complex128():
    """Test `quaxed.numpy.complex128`."""
    assert qnp.complex128(1) == jnp.complex128(1)


def test_complex64():
    """Test `quaxed.numpy.complex64`."""
    assert qnp.complex64(1) == jnp.complex64(1)


@pytest.mark.skip(reason="Not available.")
def test_complex_():
    """Test `quaxed.numpy.complex_`."""
    assert qnp.complex_(1) == jnp.complex_(1)


@pytest.mark.skip(reason="Not available.")
def test_complexfloating():
    """Test `quaxed.numpy.complexfloating`."""
    assert qnp.complexfloating(1) == jnp.complexfloating(1)


def test_compress(xbool, x1):
    """Test `quaxed.numpy.compress`."""
    assert jnp.all(qnp.compress(xbool, x1) == jnp.compress(xbool, x1))


def test_concat(x1):
    """Test `quaxed.numpy.concat`."""
    assert jnp.all(qnp.concat([x1, x1]) == jnp.concat([x1, x1]))


def test_concatenate(x1):
    """Test `quaxed.numpy.concatenate`."""
    assert jnp.all(qnp.concatenate([x1, x1]) == jnp.concatenate([x1, x1]))


def test_conj(x1):
    """Test `quaxed.numpy.conj`."""
    assert jnp.all(qnp.conj(x1) == jnp.conj(x1))


def test_conjugate(x1):
    """Test `quaxed.numpy.conjugate`."""
    assert jnp.all(qnp.conjugate(x1) == jnp.conjugate(x1))


def test_convolve(x1, x2):
    """Test `quaxed.numpy.convolve`."""
    assert jnp.all(qnp.convolve(x1, x2) == jnp.convolve(x1, x2))


def test_copy(x1):
    """Test `quaxed.numpy.copy`."""
    assert jnp.all(qnp.copy(x1) == jnp.copy(x1))


def test_copysign(x1, x2):
    """Test `quaxed.numpy.copysign`."""
    assert jnp.all(qnp.copysign(x1, x2) == jnp.copysign(x1, x2))


def test_corrcoef(x1):
    """Test `quaxed.numpy.corrcoef`."""
    assert jnp.all(qnp.corrcoef(x1) == jnp.corrcoef(x1))


def test_correlate(x1, x2):
    """Test `quaxed.numpy.correlate`."""
    assert jnp.all(qnp.correlate(x1, x2) == jnp.correlate(x1, x2))


def test_cos(x1):
    """Test `quaxed.numpy.cos`."""
    assert jnp.all(qnp.cos(x1) == jnp.cos(x1))


def test_cosh(x1):
    """Test `quaxed.numpy.cosh`."""
    assert jnp.all(qnp.cosh(x1) == jnp.cosh(x1))


def test_count_nonzero(x1):
    """Test `quaxed.numpy.count_nonzero`."""
    assert qnp.count_nonzero(x1) == jnp.count_nonzero(x1)


def test_cov(x1):
    """Test `quaxed.numpy.cov`."""
    assert jnp.all(qnp.cov(x1) == jnp.cov(x1))


def test_cross(x1, x2):
    """Test `quaxed.numpy.cross`."""
    assert jnp.all(qnp.cross(x1, x2) == jnp.cross(x1, x2))


def test_csingle():
    """Test `quaxed.numpy.csingle`."""
    assert qnp.csingle(1) == jnp.csingle(1)


def test_cumprod(x1):
    """Test `quaxed.numpy.cumprod`."""
    assert jnp.all(qnp.cumprod(x1) == jnp.cumprod(x1))


def test_cumsum(x1):
    """Test `quaxed.numpy.cumsum`."""
    assert jnp.all(qnp.cumsum(x1) == jnp.cumsum(x1))


def test_deg2rad(x1):
    """Test `quaxed.numpy.deg2rad`."""
    assert jnp.all(qnp.deg2rad(x1) == jnp.deg2rad(x1))


def test_degrees(x1):
    """Test `quaxed.numpy.degrees`."""
    assert jnp.all(qnp.degrees(x1) == jnp.degrees(x1))


def test_delete(x1):
    """Test `quaxed.numpy.delete`."""
    assert jnp.all(qnp.delete(x1, 1) == jnp.delete(x1, 1))


def test_diag(x1):
    """Test `quaxed.numpy.diag`."""
    assert jnp.all(qnp.diag(x1) == jnp.diag(x1))


def test_diag_indices():
    """Test `quaxed.numpy.diag_indices`."""
    got1, got2 = qnp.diag_indices(5)
    exp1, exp2 = jnp.diag_indices(5)
    assert jnp.all(got1 == exp1)
    assert jnp.all(got2 == exp2)


def test_diag_indices_from():
    """Test `quaxed.numpy.diag_indices_from`."""
    x = jnp.eye(4)
    got1, got2 = qnp.diag_indices_from(x)
    exp1, exp2 = jnp.diag_indices_from(x)
    assert jnp.all(got1 == exp1)
    assert jnp.all(got2 == exp2)


def test_diagflat(x1):
    """Test `quaxed.numpy.diagflat`."""
    assert jnp.all(qnp.diagflat(x1) == jnp.diagflat(x1))


def test_diagonal():
    """Test `quaxed.numpy.diagonal`."""
    x = jnp.eye(4)
    assert jnp.all(qnp.diagonal(x) == jnp.diagonal(x))


def test_diff(x1):
    """Test `quaxed.numpy.diff`."""
    assert jnp.all(qnp.diff(x1) == jnp.diff(x1))


def test_digitize(x1):
    """Test `quaxed.numpy.digitize`."""
    bins = jnp.asarray([0, 2, 4])
    assert jnp.all(qnp.digitize(x1, bins) == jnp.digitize(x1, bins))


def test_divide(x1, x2):
    """Test `quaxed.numpy.divide`."""
    assert jnp.all(qnp.divide(x1, x2) == jnp.divide(x1, x2))


def test_divmod(x1):
    """Test `quaxed.numpy.divmod`."""
    got1, got2 = qnp.divmod(x1, 2)
    exp1, exp2 = jnp.divmod(x1, 2)
    assert jnp.all(got1 == exp1)
    assert jnp.all(got2 == exp2)


def test_dot(x1, x2):
    """Test `quaxed.numpy.dot`."""
    assert jnp.all(qnp.dot(x1, x2) == jnp.dot(x1, x2))


@pytest.mark.skip(reason="no float64.")
def test_double(x1):
    """Test `quaxed.numpy.double`."""
    assert jnp.all(qnp.double(x1) == jnp.double(x1))


def test_dsplit():
    """Test `quaxed.numpy.dsplit`."""
    x = jnp.arange(16.0).reshape(2, 2, 4)
    got1, got2 = qnp.dsplit(x, 2)
    exp1, exp2 = jnp.dsplit(x, 2)
    assert jnp.all(got1 == exp1)
    assert jnp.all(got2 == exp2)


def test_dstack(x1):
    """Test `quaxed.numpy.dstack`."""
    assert jnp.all(qnp.dstack(x1) == jnp.dstack(x1))


def test_dtype(x1):
    """Test `quaxed.numpy.dtype`."""
    assert jnp.all(qnp.dtype(x1) == jnp.dtype(x1))


def test_ediff1d(x1):
    """Test `quaxed.numpy.ediff1d`."""
    assert jnp.all(qnp.ediff1d(x1) == jnp.ediff1d(x1))


def test_einsum(x1):
    """Test `quaxed.numpy.einsum`."""
    x = jnp.arange(25).reshape(5, 5)
    assert jnp.all(qnp.einsum("ii", x) == jnp.einsum("ii", x))


def test_einsum_path(x1):
    """Test `quaxed.numpy.einsum_path`."""
    key1, key2, key3 = jr.split(jr.PRNGKey(0), 3)
    a = jr.uniform(key1, (2, 2))
    b = jr.uniform(key2, (2, 5))
    c = jr.uniform(key3, (5, 2))

    got = qnp.einsum_path("ij,jk,kl->il", a, b, c, optimize="greedy")
    exp = jnp.einsum_path("ij,jk,kl->il", a, b, c, optimize="greedy")

    assert jnp.array_equal(got[0], exp[0])


def test_empty():
    """Test `quaxed.numpy.empty`."""
    assert jnp.all(qnp.empty(4) == jnp.empty(4))


def test_empty_like(x1):
    """Test `quaxed.numpy.empty_like`."""
    assert jnp.all(qnp.empty_like(x1) == jnp.empty_like(x1))


def test_equal(x1, x2):
    """Test `quaxed.numpy.equal`."""
    assert jnp.all(qnp.equal(x1, x2) == jnp.equal(x1, x2))


def test_euler_gamma(x1):
    """Test `quaxed.numpy.euler_gamma`."""
    assert qnp.euler_gamma == jnp.euler_gamma


def test_exp(x1):
    """Test `quaxed.numpy.exp`."""
    assert jnp.all(qnp.exp(x1) == jnp.exp(x1))


def test_exp2(x1):
    """Test `quaxed.numpy.exp2`."""
    assert jnp.all(qnp.exp2(x1) == jnp.exp2(x1))


def test_expand_dims(x1):
    """Test `quaxed.numpy.expand_dims`."""
    assert jnp.all(qnp.expand_dims(x1, 0) == jnp.expand_dims(x1, 0))


def test_expm1(x1):
    """Test `quaxed.numpy.expm1`."""
    assert jnp.all(qnp.expm1(x1) == jnp.expm1(x1))


def test_extract(x1):
    """Test `quaxed.numpy.extract`."""
    condition = jnp.array([True])
    assert jnp.all(qnp.extract(condition, x1) == jnp.extract(condition, x1))


def test_eye():
    """Test `quaxed.numpy.eye`."""
    assert jnp.all(qnp.eye(2) == jnp.eye(2))


def test_fabs(x1):
    """Test `quaxed.numpy.fabs`."""
    assert jnp.all(qnp.fabs(x1) == jnp.fabs(x1))


def test_fill_diagonal(x1):
    """Test `quaxed.numpy.fill_diagonal`."""
    x = jnp.eye(3)
    assert jnp.all(
        qnp.fill_diagonal(x, 2, inplace=False) == jnp.fill_diagonal(x, 2, inplace=False)
    )


def test_finfo(x1):
    """Test `quaxed.numpy.finfo`."""
    assert jnp.all(qnp.finfo(x1) == jnp.finfo(x1))


def test_fix(x1):
    """Test `quaxed.numpy.fix`."""
    assert jnp.all(qnp.fix(x1) == jnp.fix(x1))


def test_flatnonzero(x1):
    """Test `quaxed.numpy.flatnonzero`."""
    assert jnp.all(qnp.flatnonzero(x1) == jnp.flatnonzero(x1))


def test_flip(x1):
    """Test `quaxed.numpy.flip`."""
    assert jnp.all(qnp.flip(x1) == jnp.flip(x1))


def test_fliplr(x1):
    """Test `quaxed.numpy.fliplr`."""
    x = jnp.arange(4).reshape(2, 2)
    assert jnp.all(qnp.fliplr(x) == jnp.fliplr(x))


def test_flipud(x1):
    """Test `quaxed.numpy.flipud`."""
    assert jnp.all(qnp.flipud(x1) == jnp.flipud(x1))


def test_float16(x1):
    """Test `quaxed.numpy.float16`."""
    assert jnp.all(qnp.float16(x1) == jnp.float16(x1))


def test_float32(x1):
    """Test `quaxed.numpy.float32`."""
    assert jnp.all(qnp.float32(x1) == jnp.float32(x1))


@pytest.mark.skip(reason="not always available.")
def test_float64(x1):
    """Test `quaxed.numpy.float64`."""
    assert jnp.all(qnp.float64(x1) == jnp.float64(x1))


def test_float8_e4m3b11fnuz(x1):
    """Test `quaxed.numpy.float8_e4m3b11fnuz`."""
    assert jnp.all(qnp.float8_e4m3b11fnuz(x1) == jnp.float8_e4m3b11fnuz(x1))


def test_float8_e4m3fn(x1):
    """Test `quaxed.numpy.float8_e4m3fn`."""
    assert jnp.all(qnp.float8_e4m3fn(x1) == jnp.float8_e4m3fn(x1))


def test_float8_e4m3fnuz(x1):
    """Test `quaxed.numpy.float8_e4m3fnuz`."""
    assert jnp.all(qnp.float8_e4m3fnuz(x1) == jnp.float8_e4m3fnuz(x1))


def test_float8_e5m2(x1):
    """Test `quaxed.numpy.float8_e5m2`."""
    assert jnp.all(qnp.float8_e5m2(x1) == jnp.float8_e5m2(x1))


def test_float8_e5m2fnuz(x1):
    """Test `quaxed.numpy.float8_e5m2fnuz`."""
    assert jnp.all(qnp.float8_e5m2fnuz(x1) == jnp.float8_e5m2fnuz(x1))


@pytest.mark.skip(reason="not always available.")
def test_float_(x1):
    """Test `quaxed.numpy.float_`."""
    assert jnp.all(qnp.float_(x1) == jnp.float_(x1))


def test_float_power(x1, x2):
    """Test `quaxed.numpy.float_power`."""
    assert jnp.all(qnp.float_power(x1, x2) == jnp.float_power(x1, x2))


def test_floor(x1):
    """Test `quaxed.numpy.floor`."""
    assert jnp.all(qnp.floor(x1) == jnp.floor(x1))


def test_floor_divide(x1, x2):
    """Test `quaxed.numpy.floor_divide`."""
    assert jnp.all(qnp.floor_divide(x2, x1) == jnp.floor_divide(x2, x1))


def test_fmax(x1, x2):
    """Test `quaxed.numpy.fmax`."""
    assert jnp.all(qnp.fmax(x1, x2) == jnp.fmax(x1, x2))


def test_fmin(x1, x2):
    """Test `quaxed.numpy.fmin`."""
    assert jnp.all(qnp.fmin(x1, x2) == jnp.fmin(x1, x2))


def test_fmod(x1, x2):
    """Test `quaxed.numpy.fmod`."""
    assert jnp.all(qnp.fmod(x1, x2) == jnp.fmod(x1, x2))


def test_frexp(x1):
    """Test `quaxed.numpy.frexp`."""
    gotmantissa, gotexp = qnp.frexp(x1)
    expmantissa, expexp = jnp.frexp(x1)
    assert jnp.all(gotmantissa == expmantissa)
    assert jnp.all(gotexp == expexp)


def test_from_dlpack(x1):
    """Test `quaxed.numpy.from_dlpack`."""
    assert jnp.all(qnp.from_dlpack(x1) == jnp.from_dlpack(x1))


def test_frombuffer(x1):
    """Test `quaxed.numpy.frombuffer`."""
    assert jnp.all(
        qnp.frombuffer(b"\x01\x02", dtype=jnp.uint8)
        == jnp.frombuffer(b"\x01\x02", dtype=jnp.uint8)
    )


@pytest.mark.xfail(reason="JAX doesn't implement fromfile.")
def test_fromfile(x1):
    """Test `quaxed.numpy.fromfile`."""
    dt = np.dtype([("time", [("min", np.int64), ("sec", np.int64)]), ("temp", float)])
    x = np.zeros((1,), dtype=dt)
    x["time"]["min"] = 10
    x["temp"] = 98.25

    fname = tempfile.mkstemp()[1]
    x.tofile(fname)

    assert jnp.all(qnp.fromfile(x1, dtype=dt) == jnp.fromfile(x1, dtype=dt))


def test_fromfunction(x1):
    """Test `quaxed.numpy.fromfunction`."""

    def func(i, j):
        return i

    assert jnp.all(
        qnp.fromfunction(func, (2, 2), dtype=float)
        == jnp.fromfunction(func, (2, 2), dtype=float)
    )


@pytest.mark.xfail(reason="Not implemented.")
def test_fromiter():
    """Test `quaxed.numpy.fromiter`."""
    iterable = [x * x for x in range(5)]
    assert jnp.all(qnp.fromiter(iterable, float) == jnp.fromiter(iterable, float))


def test_frompyfunc(x1):
    """Test `quaxed.numpy.frompyfunc`."""
    assert jnp.all(
        qnp.frompyfunc(lambda x: x, 1, 1)(x1) == jnp.frompyfunc(lambda x: x, 1, 1)(x1)
    )


def test_fromstring():
    """Test `quaxed.numpy.fromstring`."""
    assert jnp.all(
        qnp.fromstring("1 2", dtype=int, sep=" ")
        == jnp.fromstring("1 2", dtype=int, sep=" ")
    )


def test_full():
    """Test `quaxed.numpy.full`."""
    assert jnp.all(qnp.full((2, 2), 4.0) == jnp.full((2, 2), 4.0))


def test_full_like(x1):
    """Test `quaxed.numpy.full_like`."""
    assert jnp.all(qnp.full_like(x1, 4.0) == jnp.full_like(x1, 4.0))


def test_gcd():
    """Test `quaxed.numpy.gcd`."""
    x1 = jnp.array([12, 8, 32])
    x2 = jnp.array([4, 4, 4])
    assert jnp.all(qnp.gcd(x1, x2) == jnp.gcd(x1, x2))


def test_geomspace():
    """Test `quaxed.numpy.geomspace`."""
    assert jnp.array_equal(qnp.geomspace(1.0, 100.0), jnp.geomspace(1.0, 100.0))


def test_get_printoptions():
    """Test `quaxed.numpy.get_printoptions`."""
    assert jnp.all(qnp.get_printoptions() == jnp.get_printoptions())


def test_gradient(x1):
    """Test `quaxed.numpy.gradient`."""
    assert jnp.all(qnp.gradient(x1) == jnp.gradient(x1))


def test_greater(x1, x2):
    """Test `quaxed.numpy.greater`."""
    assert jnp.all(qnp.greater(x1, x2) == jnp.greater(x1, x2))


def test_greater_equal(x1, x2):
    """Test `quaxed.numpy.greater_equal`."""
    assert jnp.all(qnp.greater_equal(x1, x2) == jnp.greater_equal(x1, x2))


def test_hamming():
    """Test `quaxed.numpy.hamming`."""
    assert jnp.all(qnp.hamming(2) == jnp.hamming(2))


def test_hanning():
    """Test `quaxed.numpy.hanning`."""
    assert jnp.all(qnp.hanning(2) == jnp.hanning(2))


def test_heaviside(x1, x2):
    """Test `quaxed.numpy.heaviside`."""
    assert jnp.all(qnp.heaviside(x1, x2) == jnp.heaviside(x1, x2))


def test_histogram(x1):
    """Test `quaxed.numpy.histogram`."""
    got = qnp.histogram(x1)
    exp = jnp.histogram(x1)
    assert all(jnp.all(g == e) for g, e in zip(got, exp, strict=True))


def test_histogram2d(x1, x2):
    """Test `quaxed.numpy.histogram2d`."""
    jnp.eye(3)
    got = qnp.histogram2d(x1, x2)
    exp = jnp.histogram2d(x1, x2)
    assert all(jnp.all(g == e) for g, e in zip(got, exp, strict=True))


def test_histogram_bin_edges(x1):
    """Test `quaxed.numpy.histogram_bin_edges`."""
    assert jnp.all(qnp.histogram_bin_edges(x1, 3) == jnp.histogram_bin_edges(x1, 3))


def test_histogramdd():
    """Test `quaxed.numpy.histogramdd`."""
    x = jnp.eye(3)
    got = qnp.histogramdd(x, bins=3)
    exp = jnp.histogramdd(x, bins=3)
    assert all(jnp.array_equal(g, e) for g, e in zip(got, exp, strict=True))


def test_hsplit(x1):
    """Test `quaxed.numpy.hsplit`."""
    got = qnp.hsplit(x1, 3)
    exp = jnp.hsplit(x1, 3)
    assert all(jnp.array_equal(g, e) for g, e in zip(got, exp, strict=True))


def test_hstack(x1):
    """Test `quaxed.numpy.hstack`."""
    assert jnp.all(qnp.hstack(x1) == jnp.hstack(x1))


def test_hypot(x1, x2):
    """Test `quaxed.numpy.hypot`."""
    assert jnp.all(qnp.hypot(x1, x2) == jnp.hypot(x1, x2))


def test_i0(x1):
    """Test `quaxed.numpy.i0`."""
    assert jnp.all(qnp.i0(x1) == jnp.i0(x1))


def test_identity():
    """Test `quaxed.numpy.identity`."""
    assert jnp.array_equal(qnp.identity(4), jnp.identity(4))


def test_iinfo(x1):
    """Test `quaxed.numpy.iinfo`."""
    x1 = x1.astype(int)
    got = qnp.iinfo(x1)
    expect = jnp.iinfo(x1)
    assert jt.tree_structure(got) == jt.tree_structure(expect)
    assert all(getattr(got, k) == getattr(expect, k) for k in ("min", "max", "dtype"))


def test_imag(x1):
    """Test `quaxed.numpy.imag`."""
    assert jnp.all(qnp.imag(x1) == jnp.imag(x1))


def test_inner(x1, x2):
    """Test `quaxed.numpy.inner`."""
    assert jnp.all(qnp.inner(x1, x2) == jnp.inner(x1, x2))


def test_insert(x1):
    """Test `quaxed.numpy.insert`."""
    assert jnp.all(qnp.insert(x1, 1, 2.0) == jnp.insert(x1, 1, 2.0))


def test_interp(x1):
    """Test `quaxed.numpy.interp`."""
    assert jnp.array_equal(qnp.interp(x1, x1, 2 * x1), jnp.interp(x1, x1, 2 * x1))


def test_intersect1d(x1):
    """Test `quaxed.numpy.intersect1d`."""
    assert jnp.array_equal(qnp.intersect1d(x1, x1), jnp.intersect1d(x1, x1))


def test_invert(x1):
    """Test `quaxed.numpy.invert`."""
    x1 = x1.astype(int)
    assert jnp.all(qnp.invert(x1) == jnp.invert(x1))


def test_isclose(x1, x2):
    """Test `quaxed.numpy.isclose`."""
    assert jnp.all(qnp.isclose(x1, x2) == jnp.isclose(x1, x2))


def test_iscomplex(x1):
    """Test `quaxed.numpy.iscomplex`."""
    assert jnp.all(qnp.iscomplex(x1) == jnp.iscomplex(x1))


def test_iscomplexobj(x1):
    """Test `quaxed.numpy.iscomplexobj`."""
    assert jnp.all(qnp.iscomplexobj(x1) == jnp.iscomplexobj(x1))


def test_isdtype(x1):
    """Test `quaxed.numpy.isdtype`."""
    assert jnp.all(qnp.isdtype(x1, "real floating") == jnp.isdtype(x1, "real floating"))


def test_isfinite(x1):
    """Test `quaxed.numpy.isfinite`."""
    assert jnp.all(qnp.isfinite(x1) == jnp.isfinite(x1))


def test_isin(x1):
    """Test `quaxed.numpy.isin`."""
    element = 2 * jnp.arange(4).reshape((2, 2))
    test_elements = jnp.asarray([1, 2, 4, 8])
    assert jnp.array_equal(
        qnp.isin(element, test_elements), jnp.isin(element, test_elements)
    )


def test_isinf(x1):
    """Test `quaxed.numpy.isinf`."""
    assert jnp.array_equal(qnp.isinf(x1), jnp.isinf(x1))


def test_isnan(x1):
    """Test `quaxed.numpy.isnan`."""
    assert jnp.array_equal(qnp.isnan(x1), jnp.isnan(x1))


def test_isneginf(x1):
    """Test `quaxed.numpy.isneginf`."""
    assert jnp.array_equal(qnp.isneginf(x1), jnp.isneginf(x1))


def test_isposinf(x1):
    """Test `quaxed.numpy.isposinf`."""
    assert jnp.array_equal(qnp.isposinf(x1), jnp.isposinf(x1))


def test_isreal(x1):
    """Test `quaxed.numpy.isreal`."""
    assert jnp.array_equal(qnp.isreal(x1), jnp.isreal(x1))


def test_isrealobj(x1):
    """Test `quaxed.numpy.isrealobj`."""
    assert jnp.array_equal(qnp.isrealobj(x1), jnp.isrealobj(x1))


def test_isscalar(x1):
    """Test `quaxed.numpy.isscalar`."""
    assert jnp.array_equal(qnp.isscalar(x1), jnp.isscalar(x1))


def test_issubdtype(x1):
    """Test `quaxed.numpy.issubdtype`."""
    assert jnp.all(
        qnp.issubdtype(x1.dtype, x1.dtype) == jnp.issubdtype(x1.dtype, x1.dtype)
    )


def test_iterable(x1):
    """Test `quaxed.numpy.iterable`."""
    assert jnp.all(qnp.iterable(x1) == jnp.iterable(x1))


def test_ix_(x1):
    """Test `quaxed.numpy.ix_`."""
    assert jnp.array_equal(qnp.ix_(x1), jnp.ix_(x1))


def test_kaiser():
    """Test `quaxed.numpy.kaiser`."""
    M = 4  # noqa: N806
    beta = jnp.asarray(3.0)
    assert jnp.array_equal(qnp.kaiser(M, beta), jnp.kaiser(M, beta))


def test_kron(x1):
    """Test `quaxed.numpy.kron`."""
    assert jnp.array_equal(qnp.kron(x1, x1), jnp.kron(x1, x1))


def test_lcm(x1, x2):
    """Test `quaxed.numpy.lcm`."""
    x1 = x1.astype(int)
    x2 = x2.astype(int)
    assert jnp.array_equal(qnp.lcm(x1, x2), jnp.lcm(x1, x2))


def test_ldexp(x1, x2):
    """Test `quaxed.numpy.ldexp`."""
    x2 = x2.astype(int)
    assert jnp.array_equal(qnp.ldexp(x1, x2), jnp.ldexp(x1, x2))


def test_left_shift(x1, x2):
    """Test `quaxed.numpy.left_shift`."""
    x1 = x1.astype(int)
    x2 = x2.astype(int)
    assert jnp.array_equal(qnp.left_shift(x1, x2), jnp.left_shift(x1, x2))


def test_less(x1, x2):
    """Test `quaxed.numpy.less`."""
    assert jnp.array_equal(qnp.less(x1, x2), jnp.less(x1, x2))


def test_less_equal(x1, x2):
    """Test `quaxed.numpy.less_equal`."""
    assert jnp.array_equal(qnp.less_equal(x1, x2), jnp.less_equal(x1, x2))


def test_lexsort(x1):
    """Test `quaxed.numpy.lexsort`."""
    assert jnp.array_equal(qnp.lexsort(x1), jnp.lexsort(x1))


def test_linspace():
    """Test `quaxed.numpy.linspace`."""
    start = 2.0
    stop = 3.0
    num = 5
    assert jnp.all(qnp.linspace(start, stop, num) == jnp.linspace(start, stop, num))


def test_load(tmp_path, x1):
    """Test `quaxed.numpy.load`."""
    path = tmp_path / "test.npy"
    jnp.save(path, x1)
    assert jnp.array_equal(qnp.load(path), jnp.load(path))


def test_log(x1):
    """Test `quaxed.numpy.log`."""
    assert jnp.array_equal(qnp.log(x1), jnp.log(x1))


def test_log10(x1):
    """Test `quaxed.numpy.log10`."""
    assert jnp.array_equal(qnp.log10(x1), jnp.log10(x1))


def test_log1p(x1):
    """Test `quaxed.numpy.log1p`."""
    assert jnp.array_equal(qnp.log1p(x1), jnp.log1p(x1))


def test_log2(x1):
    """Test `quaxed.numpy.log2`."""
    assert jnp.array_equal(qnp.log2(x1), jnp.log2(x1))


def test_logaddexp(x1, x2):
    """Test `quaxed.numpy.logaddexp`."""
    assert jnp.array_equal(qnp.logaddexp(x1, x2), jnp.logaddexp(x1, x2))


def test_logaddexp2(x1, x2):
    """Test `quaxed.numpy.logaddexp2`."""
    assert jnp.array_equal(qnp.logaddexp2(x1, x2), jnp.logaddexp2(x1, x2))


def test_logical_and(x1, x2):
    """Test `quaxed.numpy.logical_and`."""
    assert jnp.array_equal(qnp.logical_and(x1, x2), jnp.logical_and(x1, x2))


def test_logical_not(x1):
    """Test `quaxed.numpy.logical_not`."""
    assert jnp.array_equal(qnp.logical_not(x1), jnp.logical_not(x1))


def test_logical_or(x1, x2):
    """Test `quaxed.numpy.logical_or`."""
    assert jnp.array_equal(qnp.logical_or(x1, x2), jnp.logical_or(x1, x2))


def test_logical_xor(x1, x2):
    """Test `quaxed.numpy.logical_xor`."""
    assert jnp.array_equal(qnp.logical_xor(x1, x2), jnp.logical_xor(x1, x2))


def test_logspace():
    """Test `quaxed.numpy.logspace`."""
    start, stop = 0.0, 10.0
    assert jnp.array_equal(qnp.logspace(start, stop), jnp.logspace(start, stop))


def test_matmul(x1, x2):
    """Test `quaxed.numpy.matmul`."""
    assert jnp.array_equal(qnp.matmul(x1, x2), jnp.matmul(x1, x2))


def test_matrix_transpose(x1):
    """Test `quaxed.numpy.matrix_transpose`."""
    x1 = x1[..., None]  # must be 2D
    assert jnp.array_equal(qnp.matrix_transpose(x1), jnp.matrix_transpose(x1))


def test_max(x1):
    """Test `quaxed.numpy.max`."""
    assert jnp.all(qnp.max(x1) == jnp.max(x1))


def test_maximum(x1, x2):
    """Test `quaxed.numpy.maximum`."""
    assert jnp.array_equal(qnp.maximum(x1, x2), jnp.maximum(x1, x2))


def test_mean(x1):
    """Test `quaxed.numpy.mean`."""
    assert qnp.mean(x1) == jnp.mean(x1)


def test_median(x1):
    """Test `quaxed.numpy.median`."""
    assert qnp.median(x1) == jnp.median(x1)


def test_meshgrid(x1, x2):
    """Test `quaxed.numpy.meshgrid`."""
    assert jnp.array_equal(qnp.meshgrid(x1, x2), jnp.meshgrid(x1, x2))


def test_min(x1):
    """Test `quaxed.numpy.min`."""
    assert qnp.min(x1) == jnp.min(x1)


def test_minimum(x1, x2):
    """Test `quaxed.numpy.minimum`."""
    assert jnp.array_equal(qnp.minimum(x1, x2), jnp.minimum(x1, x2))


def test_mod(x1):
    """Test `quaxed.numpy.mod`."""
    assert jnp.array_equal(qnp.mod(x1, 2), jnp.mod(x1, 2))


def test_modf(x1):
    """Test `quaxed.numpy.modf`."""
    got_frac, got_intg = qnp.modf(x1)
    exp_frac, exp_intg = jnp.modf(x1)
    assert jnp.array_equal(got_frac, exp_frac)
    assert jnp.array_equal(got_intg, exp_intg)


def test_moveaxis(x1):
    """Test `quaxed.numpy.moveaxis`."""
    assert jnp.array_equal(qnp.moveaxis(x1[None], 0, 1), jnp.moveaxis(x1[None], 0, 1))


def test_multiply(x1, x2):
    """Test `quaxed.numpy.multiply`."""
    assert jnp.array_equal(qnp.multiply(x1, x2), jnp.multiply(x1, x2))


def test_nan_to_num(x1):
    """Test `quaxed.numpy.nan_to_num`."""
    assert jnp.all(qnp.nan_to_num(x1) == jnp.nan_to_num(x1))


def test_nanargmax(x1):
    """Test `quaxed.numpy.nanargmax`."""
    assert jnp.all(qnp.nanargmax(x1) == jnp.nanargmax(x1))


def test_nanargmin(x1):
    """Test `quaxed.numpy.nanargmin`."""
    assert jnp.all(qnp.nanargmin(x1) == jnp.nanargmin(x1))


def test_nancumprod(x1):
    """Test `quaxed.numpy.nancumprod`."""
    assert jnp.all(qnp.nancumprod(x1) == jnp.nancumprod(x1))


def test_nancumsum(x1):
    """Test `quaxed.numpy.nancumsum`."""
    assert jnp.all(qnp.nancumsum(x1) == jnp.nancumsum(x1))


def test_nanmax(x1):
    """Test `quaxed.numpy.nanmax`."""
    assert jnp.all(qnp.nanmax(x1) == jnp.nanmax(x1))


def test_nanmean(x1):
    """Test `quaxed.numpy.nanmean`."""
    assert jnp.all(qnp.nanmean(x1) == jnp.nanmean(x1))


def test_nanmedian(x1):
    """Test `quaxed.numpy.nanmedian`."""
    assert jnp.all(qnp.nanmedian(x1) == jnp.nanmedian(x1))


def test_nanmin(x1):
    """Test `quaxed.numpy.nanmin`."""
    assert jnp.all(qnp.nanmin(x1) == jnp.nanmin(x1))


def test_nanpercentile(x1):
    """Test `quaxed.numpy.nanpercentile`."""
    assert jnp.array_equal(qnp.nanpercentile(x1, 50), jnp.nanpercentile(x1, 50))


def test_nanprod(x1):
    """Test `quaxed.numpy.nanprod`."""
    assert jnp.array_equal(qnp.nanprod(x1), jnp.nanprod(x1))


def test_nanquantile(x1):
    """Test `quaxed.numpy.nanquantile`."""
    assert jnp.array_equal(qnp.nanquantile(x1, 0.5), jnp.nanquantile(x1, 0.5))


def test_nanstd(x1):
    """Test `quaxed.numpy.nanstd`."""
    assert qnp.nanstd(x1) == jnp.nanstd(x1)


def test_nansum(x1):
    """Test `quaxed.numpy.nansum`."""
    assert qnp.nansum(x1) == jnp.nansum(x1)


def test_nanvar(x1):
    """Test `quaxed.numpy.nanvar`."""
    assert qnp.nanvar(x1) == jnp.nanvar(x1)


def test_ndim(x1):
    """Test `quaxed.numpy.ndim`."""
    assert qnp.ndim(x1) == jnp.ndim(x1)


def test_negative(x1):
    """Test `quaxed.numpy.negative`."""
    assert jnp.array_equal(qnp.negative(x1), jnp.negative(x1))


def test_nextafter(x1, x2):
    """Test `quaxed.numpy.nextafter`."""
    assert jnp.array_equal(qnp.nextafter(x1, x2), jnp.nextafter(x1, x2))


def test_nonzero(x1):
    """Test `quaxed.numpy.nonzero`."""
    assert jnp.array_equal(qnp.nonzero(x1), jnp.nonzero(x1))


def test_not_equal(x1, x2):
    """Test `quaxed.numpy.not_equal`."""
    assert jnp.array_equal(qnp.not_equal(x1, x2), jnp.not_equal(x1, x2))


def test_ones():
    """Test `quaxed.numpy.ones`."""
    assert jnp.array_equal(qnp.ones(5), jnp.ones(5))


def test_ones_like(x1):
    """Test `quaxed.numpy.ones_like`."""
    assert jnp.array_equal(qnp.ones_like(x1), jnp.ones_like(x1))


def test_outer(x1, x2):
    """Test `quaxed.numpy.outer`."""
    assert jnp.array_equal(qnp.outer(x1, x2), jnp.outer(x1, x2))


def test_packbits():
    """Test `quaxed.numpy.packbits`."""
    x = jnp.array([[[1, 0, 1], [0, 1, 0]]], dtype=jnp.uint8)
    assert jnp.array_equal(qnp.packbits(x), jnp.packbits(x))


def test_pad(x1):
    """Test `quaxed.numpy.pad`."""
    assert jnp.array_equal(qnp.pad(x1, 20), jnp.pad(x1, 20))


def test_partition(x1):
    """Test `quaxed.numpy.partition`."""
    assert jnp.all(qnp.partition(x1, 1) == jnp.partition(x1, 1))


def test_percentile(x1):
    """Test `quaxed.numpy.percentile`."""
    assert qnp.percentile(x1, 50) == jnp.percentile(x1, 50)


def test_permute_dims(x1):
    """Test `quaxed.numpy.permute_dims`."""
    x1 = x1[None]
    assert jnp.all(qnp.permute_dims(x1, (0, 1)) == jnp.permute_dims(x1, (0, 1)))


def test_piecewise(x1):
    """Test `quaxed.numpy.piecewise`."""
    got = qnp.piecewise(x1, [x1 < 0, x1 >= 0], [-1, 1])
    exp = jnp.piecewise(x1, [x1 < 0, x1 >= 0], [-1, 1])
    assert jnp.array_equal(got, exp)


def test_place(x1):
    """Test `quaxed.numpy.place`."""
    mask = x1 > qnp.mean(x1)
    assert jnp.array_equal(
        qnp.place(x1, mask, 0, inplace=False), jnp.place(x1, mask, 0, inplace=False)
    )


def test_poly(x1):
    """Test `quaxed.numpy.poly`."""
    assert jnp.all(qnp.poly(x1) == jnp.poly(x1))


@pytest.mark.skip("TODO")
def test_polyadd(x1):
    """Test `quaxed.numpy.polyadd`."""
    assert jnp.all(qnp.polyadd(x1) == jnp.polyadd(x1))


@pytest.mark.skip("TODO")
def test_polyder(x1):
    """Test `quaxed.numpy.polyder`."""
    assert jnp.all(qnp.polyder(x1) == jnp.polyder(x1))


@pytest.mark.skip("TODO")
def test_polydiv(x1):
    """Test `quaxed.numpy.polydiv`."""
    assert jnp.all(qnp.polydiv(x1) == jnp.polydiv(x1))


@pytest.mark.skip("TODO")
def test_polyfit(x1):
    """Test `quaxed.numpy.polyfit`."""
    assert jnp.all(qnp.polyfit(x1) == jnp.polyfit(x1))


@pytest.mark.skip("TODO")
def test_polyint(x1):
    """Test `quaxed.numpy.polyint`."""
    assert jnp.all(qnp.polyint(x1) == jnp.polyint(x1))


@pytest.mark.skip("TODO")
def test_polymul(x1):
    """Test `quaxed.numpy.polymul`."""
    assert jnp.all(qnp.polymul(x1) == jnp.polymul(x1))


@pytest.mark.skip("TODO")
def test_polysub(x1):
    """Test `quaxed.numpy.polysub`."""
    assert jnp.all(qnp.polysub(x1) == jnp.polysub(x1))


@pytest.mark.skip("TODO")
def test_polyval(x1):
    """Test `quaxed.numpy.polyval`."""
    assert jnp.all(qnp.polyval(x1) == jnp.polyval(x1))


def test_positive(x1):
    """Test `quaxed.numpy.positive`."""
    assert jnp.array_equal(qnp.positive(x1), jnp.positive(x1))


def test_pow(x1):
    """Test `quaxed.numpy.pow`."""
    assert jnp.array_equal(qnp.pow(x1, 2), jnp.pow(x1, 2))


def test_power(x1):
    """Test `quaxed.numpy.power`."""
    assert jnp.array_equal(qnp.power(x1, 2.5), jnp.power(x1, 2.5))


def test_prod(x1):
    """Test `quaxed.numpy.prod`."""
    assert qnp.prod(x1) == jnp.prod(x1)


def test_ptp(x1):
    """Test `quaxed.numpy.ptp`."""
    assert jnp.all(qnp.ptp(x1) == jnp.ptp(x1))


@pytest.mark.skip("TODO")
def test_put(x1):
    """Test `quaxed.numpy.put`."""
    assert jnp.all(qnp.put(x1) == jnp.put(x1))


def test_quantile(x1):
    """Test `quaxed.numpy.quantile`."""
    assert qnp.quantile(x1, 0.5) == jnp.quantile(x1, 0.5)


def test_r_(x1, x2):
    """Test `quaxed.numpy.r_`."""
    assert jnp.all(qnp.r_[x1, x2] == jnp.r_[x1, x2])


def test_rad2deg(x1):
    """Test `quaxed.numpy.rad2deg`."""
    assert jnp.array_equal(qnp.rad2deg(x1), jnp.rad2deg(x1))


def test_radians(x1):
    """Test `quaxed.numpy.radians`."""
    assert jnp.all(qnp.radians(x1) == jnp.radians(x1))


def test_ravel(x1):
    """Test `quaxed.numpy.ravel`."""
    assert jnp.all(qnp.ravel(x1) == jnp.ravel(x1))


@pytest.mark.skip("TODO")
def test_ravel_multi_index(x1):
    """Test `quaxed.numpy.ravel_multi_index`."""
    assert jnp.all(qnp.ravel_multi_index(x1) == jnp.ravel_multi_index(x1))


def test_real(x1):
    """Test `quaxed.numpy.real`."""
    assert jnp.array_equal(qnp.real(x1), jnp.real(x1))


def test_reciprocal(x1):
    """Test `quaxed.numpy.reciprocal`."""
    assert jnp.array_equal(qnp.reciprocal(x1), jnp.reciprocal(x1))


def test_remainder(x1, x2):
    """Test `quaxed.numpy.remainder`."""
    assert jnp.all(qnp.remainder(x1, x2) == jnp.remainder(x1, x2))


def test_repeat(x1):
    """Test `quaxed.numpy.repeat`."""
    assert jnp.array_equal(qnp.repeat(x1, 3), jnp.repeat(x1, 3))


def test_reshape(x1):
    """Test `quaxed.numpy.reshape`."""
    assert jnp.array_equal(qnp.reshape(x1, (1, -1)), jnp.reshape(x1, (1, -1)))


def test_resize(x1):
    """Test `quaxed.numpy.resize`."""
    assert jnp.array_equal(qnp.resize(x1, (1, len(x1))), jnp.resize(x1, (1, len(x1))))


def test_result_type(x1):
    """Test `quaxed.numpy.result_type`."""
    assert qnp.result_type(3, x1) == jnp.result_type(2, x1)


def test_right_shift(xbool):
    """Test `quaxed.numpy.right_shift`."""
    assert jnp.array_equal(qnp.right_shift(xbool, xbool), jnp.right_shift(xbool, xbool))


def test_rint(x1):
    """Test `quaxed.numpy.rint`."""
    assert jnp.array_equal(qnp.rint(x1), jnp.rint(x1))


def test_roll(x1):
    """Test `quaxed.numpy.roll`."""
    assert jnp.array_equal(qnp.roll(x1, 4), jnp.roll(x1, 4))


def test_rollaxis(x1):
    """Test `quaxed.numpy.rollaxis`."""
    assert jnp.array_equal(qnp.rollaxis(x1, -1), jnp.rollaxis(x1, -1))


def test_roots(x1):
    """Test `quaxed.numpy.roots`."""
    assert jnp.array_equal(qnp.roots(x1), jnp.roots(x1))


def test_rot90(x1):
    """Test `quaxed.numpy.rot90`."""
    x1 = x1[None]
    assert jnp.array_equal(qnp.rot90(x1), jnp.rot90(x1))


def test_round(x1):
    """Test `quaxed.numpy.round`."""
    assert jnp.all(qnp.round(x1) == jnp.round(x1))


def test_round_(x1):
    """Test `quaxed.numpy.round_`."""
    with pytest.deprecated_call():
        assert jnp.all(qnp.round_(x1) == jnp.round_(x1))


@pytest.mark.skip("TODO")
def test_save(x1):
    """Test `quaxed.numpy.save`."""
    assert jnp.all(qnp.save(x1) == jnp.save(x1))


@pytest.mark.skip("TODO")
def test_savez(x1):
    """Test `quaxed.numpy.savez`."""
    assert jnp.all(qnp.savez(x1) == jnp.savez(x1))


@pytest.mark.skip("TODO")
def test_searchsorted(x1):
    """Test `quaxed.numpy.searchsorted`."""
    assert jnp.all(qnp.searchsorted(x1) == jnp.searchsorted(x1))


@pytest.mark.skip("TODO")
def test_select(x1):
    """Test `quaxed.numpy.select`."""
    assert jnp.all(qnp.select(x1) == jnp.select(x1))


def test_setdiff1d(x1, x2):
    """Test `quaxed.numpy.setdiff1d`."""
    assert jnp.array_equal(qnp.setdiff1d(x1, x2), jnp.setdiff1d(x1, x2))


def test_setxor1d(x1, x2):
    """Test `quaxed.numpy.setxor1d`."""
    assert jnp.array_equal(qnp.setxor1d(x1, x2), jnp.setxor1d(x1, x2))


def test_shape(x1):
    """Test `quaxed.numpy.shape`."""
    assert qnp.shape(x1) == jnp.shape(x1)


def test_sign(x1):
    """Test `quaxed.numpy.sign`."""
    assert jnp.array_equal(qnp.sign(x1), jnp.sign(x1))


def test_signbit(x1):
    """Test `quaxed.numpy.signbit`."""
    assert jnp.all(qnp.signbit(x1) == jnp.signbit(x1))


def test_sin(x1):
    """Test `quaxed.numpy.sin`."""
    assert jnp.array_equal(qnp.sin(x1), jnp.sin(x1))


def test_sinc(x1):
    """Test `quaxed.numpy.sinc`."""
    assert jnp.array_equal(qnp.sinc(x1), jnp.sinc(x1))


def test_sinh(x1):
    """Test `quaxed.numpy.sinh`."""
    assert jnp.array_equal(qnp.sinh(x1), jnp.sinh(x1))


def test_size(x1):
    """Test `quaxed.numpy.size`."""
    assert qnp.size(x1) == jnp.size(x1)


def test_sort(x1):
    """Test `quaxed.numpy.sort`."""
    assert jnp.array_equal(qnp.sort(x1), jnp.sort(x1))


def test_sort_complex(x1):
    """Test `quaxed.numpy.sort_complex`."""
    assert jnp.array_equal(qnp.sort_complex(x1), jnp.sort_complex(x1))


def test_split(x1):
    """Test `quaxed.numpy.split`."""
    got = qnp.split(x1, 3)
    exp = qnp.split(x1, 3)
    assert all(jnp.array_equal(g, e) for g, e in zip(got, exp, strict=True))


def test_sqrt(x1):
    """Test `quaxed.numpy.sqrt`."""
    assert jnp.array_equal(qnp.sqrt(x1), jnp.sqrt(x1))


def test_square(x1):
    """Test `quaxed.numpy.square`."""
    assert jnp.array_equal(qnp.square(x1), jnp.square(x1))


def test_squeeze(x1):
    """Test `quaxed.numpy.squeeze`."""
    x1 = x1[None]
    assert jnp.array_equal(qnp.squeeze(x1), jnp.squeeze(x1))


def test_stack(x1, x2):
    """Test `quaxed.numpy.stack`."""
    assert jnp.array_equal(qnp.stack((x1, x2)), jnp.stack((x1, x2)))


def test_std(x1):
    """Test `quaxed.numpy.std`."""
    assert qnp.std(x1) == jnp.std(x1)


def test_subtract(x1, x2):
    """Test `quaxed.numpy.subtract`."""
    assert jnp.array_equal(qnp.subtract(x1, x2), jnp.subtract(x1, x2))


def test_sum(x1):
    """Test `quaxed.numpy.sum`."""
    assert qnp.sum(x1) == jnp.sum(x1)


def test_swapaxes(x1):
    """Test `quaxed.numpy.swapaxes`."""
    x1 = x1[None]
    assert jnp.all(qnp.swapaxes(x1, 0, 1) == jnp.swapaxes(x1, 0, 1))


@pytest.mark.skip("TODO")
def test_take(x1):
    """Test `quaxed.numpy.take`."""
    assert jnp.all(qnp.take(x1) == jnp.take(x1))


@pytest.mark.skip("TODO")
def test_take_along_axis(x1):
    """Test `quaxed.numpy.take_along_axis`."""
    assert jnp.all(qnp.take_along_axis(x1) == jnp.take_along_axis(x1))


def test_tan(x1):
    """Test `quaxed.numpy.tan`."""
    assert jnp.array_equal(qnp.tan(x1), jnp.tan(x1))


def test_tanh(x1):
    """Test `quaxed.numpy.tanh`."""
    assert jnp.array_equal(qnp.tanh(x1), jnp.tanh(x1))


@pytest.mark.skip("TODO")
def test_tensordot(x1, x2):
    """Test `quaxed.numpy.tensordot`."""
    assert jnp.array_equal(qnp.tensordot(x1, x2), jnp.tensordot(x1, x2))


def test_tile(x1):
    """Test `quaxed.numpy.tile`."""
    assert jnp.all(qnp.tile(x1, 3) == jnp.tile(x1, 3))


def test_trace(x1):
    """Test `quaxed.numpy.trace`."""
    x1 = x1[None]
    assert jnp.array_equal(qnp.trace(x1), jnp.trace(x1))


def test_transpose(x1):
    """Test `quaxed.numpy.transpose`."""
    x1 = x1[None]
    assert jnp.array_equal(qnp.transpose(x1), jnp.transpose(x1))


def test_tril():
    """Test `quaxed.numpy.tril`."""
    x = jnp.eye(4)
    assert jnp.array_equal(qnp.tril(x), jnp.tril(x))


def test_tril_indices_from():
    """Test `quaxed.numpy.tril_indices_from`."""
    x = jnp.eye(4)
    assert jnp.array_equal(qnp.tril_indices_from(x), jnp.tril_indices_from(x))


def test_trim_zeros(x1):
    """Test `quaxed.numpy.trim_zeros`."""
    x = jnp.concatenate((np.asarray([0.0, 0, 0]), x1, jnp.asarray([0.0, 0, 0])))
    assert jnp.array_equal(qnp.trim_zeros(x), jnp.trim_zeros(x))


def test_triu(x1):
    """Test `quaxed.numpy.triu`."""
    x = jnp.eye(4)
    assert jnp.array_equal(qnp.triu(x), jnp.triu(x))


def test_triu_indices_from():
    """Test `quaxed.numpy.triu_indices_from`."""
    x = jnp.eye(4)
    assert jnp.array_equal(qnp.triu_indices_from(x), jnp.triu_indices_from(x))


def test_true_divide(x1, x2):
    """Test `quaxed.numpy.true_divide`."""
    assert jnp.array_equal(qnp.true_divide(x1, x2), jnp.true_divide(x1, x2))


def test_trunc(x1):
    """Test `quaxed.numpy.trunc`."""
    assert jnp.array_equal(qnp.trunc(x1), jnp.trunc(x1))


@pytest.mark.skip("TODO")
def test_ufunc(x1):
    """Test `quaxed.numpy.ufunc`."""
    assert jnp.all(qnp.ufunc(x1) == jnp.ufunc(x1))


def test_union1d(x1, x2):
    """Test `quaxed.numpy.union1d`."""
    assert jnp.array_equal(qnp.union1d(x1, x2), jnp.union1d(x1, x2))


def test_unique(x1):
    """Test `quaxed.numpy.unique`."""
    assert jnp.array_equal(qnp.unique(x1), jnp.unique(x1))


def test_unique_all(x1):
    """Test `quaxed.numpy.unique_all`."""
    got = qnp.unique_all(x1)
    exp = jnp.unique_all(x1)
    assert all(jnp.array_equal(g, e) for g, e in zip(got, exp, strict=True))


def test_unique_counts(x1):
    """Test `quaxed.numpy.unique_counts`."""
    assert jnp.array_equal(qnp.unique_counts(x1), jnp.unique_counts(x1))


def test_unique_inverse(x1):
    """Test `quaxed.numpy.unique_inverse`."""
    assert jnp.array_equal(qnp.unique_inverse(x1), jnp.unique_inverse(x1))


def test_unique_values(x1):
    """Test `quaxed.numpy.unique_values`."""
    assert jnp.array_equal(qnp.unique_values(x1), jnp.unique_values(x1))


@pytest.mark.skip("TODO")
def test_unpackbits(x1):
    """Test `quaxed.numpy.unpackbits`."""
    assert jnp.all(qnp.unpackbits(x1) == jnp.unpackbits(x1))


@pytest.mark.skip("TODO")
def test_unravel_index(x1):
    """Test `quaxed.numpy.unravel_index`."""
    assert jnp.all(qnp.unravel_index(x1) == jnp.unravel_index(x1))


def test_unwrap(x1):
    """Test `quaxed.numpy.unwrap`."""
    assert jnp.array_equal(qnp.unwrap(x1), jnp.unwrap(x1))


@pytest.mark.skip("TODO")
def test_vander(x1):
    """Test `quaxed.numpy.vander`."""
    assert jnp.all(qnp.vander(x1) == jnp.vander(x1))


def test_var(x1):
    """Test `quaxed.numpy.var`."""
    assert qnp.var(x1) == jnp.var(x1)


def test_vdot(x1, x2):
    """Test `quaxed.numpy.vdot`."""
    assert jnp.all(qnp.vdot(x1, x2) == jnp.vdot(x1, x2))


def test_vecdot(x1, x2):
    """Test `quaxed.numpy.vecdot`."""
    assert jnp.all(qnp.vecdot(x1, x2) == jnp.vecdot(x1, x2))


def test_vectorize(x1):
    """Test `quaxed.numpy.vectorize`."""

    @qnp.vectorize
    def f(x):
        return x + 1

    assert jnp.all(f(x1) == jnp.vectorize(lambda x: x + 1)(x1))


def test_vsplit(x1):
    """Test `quaxed.numpy.vsplit`."""
    got = qnp.vsplit(x1, 3)
    exp = jnp.vsplit(x1, 3)
    assert all(jnp.array_equal(g, e) for g, e in zip(got, exp, strict=True))


def test_vstack(x1, x2):
    """Test `quaxed.numpy.vstack`."""
    assert jnp.array_equal(qnp.vstack((x1, x2)), jnp.vstack((x1, x2)))


def test_where(x1, x2):
    """Test `quaxed.numpy.where`."""
    cond = jnp.ones_like(x1, dtype=bool)
    assert jnp.array_equal(qnp.where(cond, x1, x2), jnp.where(cond, x1, x2))


def test_zeros(x1):
    """Test `quaxed.numpy.zeros`."""
    assert jnp.array_equal(qnp.zeros(4), jnp.zeros(4))


def test_zeros_like(x1):
    """Test `quaxed.numpy.zeros_like`."""
    assert jnp.array_equal(qnp.zeros_like(x1), jnp.zeros_like(x1))



"""Quaxed :mod:`jax.lax`."""
# pylint: disable=undefined-all-variable

__all__ = [  # noqa: F822
    "cholesky",
    "eig",
    "eigh",
    "hessenberg",
    "lu",
    "householder_product",
    "qdwh",
    "qr",
    "schur",
    "svd",
    "triangular_solve",
    "tridiagonal",
    "tridiagonal_solve",
]


import sys
from collections.abc import Callable
from typing import Any

from jax import lax
from quax import quaxify


def __dir__() -> list[str]:
    """List the module contents."""
    return sorted(__all__)


# TODO: return type hint signature
def __getattr__(name: str) -> Callable[..., Any]:
    """Get the :external:`quax.quaxify`'ed function."""
    if name not in __all__:
        msg = f"Cannot get {name} from quaxed.lax.linalg"
        raise AttributeError(msg)

    # Quaxify the operator
    out = quaxify(getattr(lax.linalg, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out



"""Quaxed :mod:`jax.numpy`."""

__all__ = ["vectorize"]

import functools
from collections.abc import Callable, Collection
from typing import Any, TypeVar

import equinox as eqx
import jax
from jax._src.numpy.vectorize import (
    _apply_excluded,
    _check_output_dims,
    _parse_gufunc_signature,
    _parse_input_dimensions,
)

from ._core import asarray, squeeze
from ._core import expand_dims as _expand_dims

T = TypeVar("T")


def expand_dims(a: T, axis: int | tuple[int, ...]) -> T:
    dynamic, static = eqx.partition(a, eqx.is_array_like)
    expanded_dynamic = jax.tree.map(lambda x: _expand_dims(x, axis), dynamic)
    return eqx.combine(expanded_dynamic, static)


def vectorize(  # noqa: C901
    pyfunc: Callable[..., Any],
    *,
    excluded: Collection[int | str] = frozenset(),
    signature: str | None = None,
) -> Callable[..., Any]:
    """Define a vectorized function with broadcasting.

    This is a copy-paste from :func:`jax.numpy.vectorize`, but the internals are
    all replaced with their :mod:`quaxed` counterparts to allow quax-friendly
    objects to pass through. The only thing that isn't quaxed is `jax.vmap`,
    which allows any array-like object to pass through without converting it.
    Note that this behaviour is DIFFERENT than doing ``quaxify(jnp.vectorize)``
    since `quaxify` makes objects look like arrays, not their actual type,
    which can be problematic. This function passes through the objects
    unchanged (so long as they are amenable to the reshapes and ``vamap``).

    Arguments:
    ---------
    pyfunc: callable
        function to vectorize.
    excluded: Collection[int | str], optional
        optional set of integers representing positional arguments for which the
        function will not be vectorized. These will be passed directly to
        ``pyfunc`` unmodified.
    signature: str | None
        optional generalized universal function signature, e.g.,
        ``(m,n),(n)->(m)`` for vectorized matrix-vector multiplication. If
        provided, ``pyfunc`` will be called with (and expected to return) arrays
        with shapes given by the size of corresponding core dimensions. By
        default, pyfunc is assumed to take scalars arrays as input and output.

    Returns
    -------
    callable
        Vectorized version of the given function.

    Here are a few examples of how one could write vectorized linear algebra
    routines using :func:`vectorize`:

    >>> from functools import partial
    >>> import quaxed.numpy as jnp

    >>> @partial(jnp.vectorize, signature='(k),(k)->(k)')
    ... def cross_product(a, b):
    ...   assert a.shape == b.shape and a.ndim == b.ndim == 1
    ...   return jnp.array([a[1] * b[2] - a[2] * b[1],
    ...                     a[2] * b[0] - a[0] * b[2],
    ...                     a[0] * b[1] - a[1] * b[0]])

    >>> @partial(jnp.vectorize, signature='(n,m),(m)->(n)')
    ... def matrix_vector_product(matrix, vector):
    ...   assert matrix.ndim == 2 and matrix.shape[1:] == vector.shape
    ...   return matrix @ vector

    These functions are only written to handle 1D or 2D arrays (the ``assert``
    statements will never be violated), but with vectorize they support
    arbitrary dimensional inputs with NumPy style broadcasting, e.g.,

    >>> cross_product(jnp.ones(3), jnp.ones(3)).shape
    (3,)
    >>> cross_product(jnp.ones((2, 3)), jnp.ones(3)).shape
    (2, 3)
    >>> cross_product(jnp.ones((1, 2, 3)), jnp.ones((2, 1, 3))).shape
    (2, 2, 3)
    >>> matrix_vector_product(jnp.ones(3), jnp.ones(3))  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: input with shape (3,) does not have enough dimensions for all
    core dimensions ('n', 'k') on vectorized function with excluded=frozenset()
    and signature='(n,k),(k)->(k)'
    >>> matrix_vector_product(jnp.ones((2, 3)), jnp.ones(3)).shape
    (2,)
    >>> matrix_vector_product(jnp.ones((2, 3)), jnp.ones((4, 3))).shape
    (4, 2)

    Note that this has different semantics than `jnp.matmul`:

    >>> jnp.matmul(jnp.ones((2, 3)), jnp.ones((4, 3)))  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    TypeError: dot_general requires contracting dimensions to have the same shape, got [3] and [4].
    """  # noqa: E501
    if any(not isinstance(exclude, str | int) for exclude in excluded):
        msg = (
            "jax.numpy.vectorize can only exclude integer or string arguments, "
            f"but excluded={excluded!r}"
        )
        raise TypeError(msg)

    if any(isinstance(e, int) and e < 0 for e in excluded):
        msg = f"excluded={excluded!r} contains negative numbers"
        raise ValueError(msg)

    @functools.wraps(pyfunc)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        error_context = (
            f"on vectorized function with excluded={excluded!r} and "
            f"signature={signature!r}"
        )
        excluded_func, args, kwargs = _apply_excluded(pyfunc, excluded, args, kwargs)

        if signature is not None:
            input_core_dims, output_core_dims = _parse_gufunc_signature(signature)
        else:
            input_core_dims = [()] * len(args)
            output_core_dims = None

        none_args = {i for i, arg in enumerate(args) if arg is None}
        if any(none_args):
            if any(input_core_dims[i] != () for i in none_args):
                msg = f"Cannot pass None at locations {none_args} with {signature=}"
                raise ValueError(msg)
            excluded_func, args, _ = _apply_excluded(excluded_func, none_args, args, {})
            input_core_dims = [
                dim for i, dim in enumerate(input_core_dims) if i not in none_args
            ]

        args = tuple(map(asarray, args))

        broadcast_shape, dim_sizes = _parse_input_dimensions(
            args, input_core_dims, error_context
        )

        checked_func = _check_output_dims(
            excluded_func, dim_sizes, output_core_dims, error_context
        )

        # Rather than broadcasting all arguments to full broadcast shapes, prefer
        # expanding dimensions using vmap. By pushing broadcasting
        # into vmap, we can make use of more efficient batching rules for
        # primitives where only some arguments are batched (e.g., for
        # lax_linalg.triangular_solve), and avoid instantiating large broadcasted
        # arrays.

        squeezed_args = []
        rev_filled_shapes = []

        for arg, core_dims in zip(args, input_core_dims, strict=True):
            noncore_shape = arg.shape[: arg.ndim - len(core_dims)]

            pad_ndim = len(broadcast_shape) - len(noncore_shape)
            filled_shape = pad_ndim * (1,) + noncore_shape
            rev_filled_shapes.append(filled_shape[::-1])

            squeeze_indices = tuple(
                i for i, size in enumerate(noncore_shape) if size == 1
            )
            squeezed_arg = squeeze(arg, axis=squeeze_indices)
            squeezed_args.append(squeezed_arg)

        vectorized_func = checked_func
        dims_to_expand = []
        for negdim, axis_sizes in enumerate(zip(*rev_filled_shapes, strict=True)):
            in_axes = tuple(None if size == 1 else 0 for size in axis_sizes)
            if all(axis is None for axis in in_axes):
                dims_to_expand.append(len(broadcast_shape) - 1 - negdim)
            else:
                vectorized_func = jax.vmap(vectorized_func, in_axes)
        result = vectorized_func(*squeezed_args)

        if not dims_to_expand:
            out = result
        elif isinstance(result, tuple):
            dims_to_expand_ = tuple(dims_to_expand)
            out = tuple(expand_dims(r, axis=dims_to_expand_) for r in result)
        else:
            out = expand_dims(result, axis=tuple(dims_to_expand))

        return out

    return wrapped



import sys
from operator import abs as abs
from operator import add as add
from operator import and_ as and_
from operator import attrgetter as attrgetter
from operator import concat as concat
from operator import contains as contains
from operator import countOf as countOf
from operator import delitem as delitem
from operator import eq as eq
from operator import floordiv as floordiv
from operator import ge as ge
from operator import getitem as getitem
from operator import gt as gt
from operator import iadd as iadd
from operator import iand as iand
from operator import iconcat as iconcat
from operator import ifloordiv as ifloordiv
from operator import ilshift as ilshift
from operator import imatmul as imatmul
from operator import imod as imod
from operator import imul as imul
from operator import index as index
from operator import indexOf as indexOf
from operator import inv as inv
from operator import invert as invert
from operator import ior as ior
from operator import ipow as ipow
from operator import irshift as irshift
from operator import is_ as is_
from operator import is_not as is_not
from operator import isub as isub
from operator import itemgetter as itemgetter
from operator import itruediv as itruediv
from operator import ixor as ixor
from operator import le as le
from operator import length_hint as length_hint
from operator import lshift as lshift
from operator import lt as lt
from operator import matmul as matmul
from operator import methodcaller as methodcaller
from operator import mod as mod
from operator import mul as mul
from operator import ne as ne
from operator import neg as neg
from operator import not_ as not_
from operator import or_ as or_
from operator import pos as pos
from operator import pow as pow
from operator import rshift as rshift
from operator import setitem as setitem
from operator import sub as sub
from operator import truediv as truediv
from operator import truth as truth
from operator import xor as xor

if sys.version_info >= (3, 11):
    from operator import call as call



"""Quaxed :mod:`jax.lax`."""
# pylint: disable=undefined-all-variable

__all__ = [
    # ----- Operators -----
    "abs",
    "acos",
    "acosh",
    "add",
    # "after_all",
    "approx_max_k",
    "approx_min_k",
    "argmax",
    "argmin",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "batch_matmul",
    "bessel_i0e",
    "bessel_i1e",
    "betainc",
    "bitcast_convert_type",
    "bitwise_and",
    "bitwise_not",
    "bitwise_or",
    "bitwise_xor",
    "population_count",
    "broadcast",
    "broadcast_in_dim",
    "broadcast_shapes",
    "broadcast_to_rank",
    "broadcasted_iota",
    "cbrt",
    "ceil",
    "clamp",
    "clz",
    "collapse",
    "complex",
    "concatenate",
    "conj",
    "conv",
    "convert_element_type",
    "conv_dimension_numbers",
    "conv_general_dilated",
    "conv_general_dilated_local",
    "conv_general_dilated_patches",
    "conv_transpose",
    "conv_with_general_padding",
    "cos",
    "cosh",
    "cumlogsumexp",
    "cummax",
    "cummin",
    "cumprod",
    "cumsum",
    "digamma",
    "div",
    "dot",
    "dot_general",
    "dynamic_index_in_dim",
    "dynamic_slice",
    "dynamic_slice_in_dim",
    "dynamic_update_index_in_dim",
    "dynamic_update_slice",
    "dynamic_update_slice_in_dim",
    "eq",
    "erf",
    "erfc",
    "erf_inv",
    "exp",
    "expand_dims",
    "expm1",
    "fft",
    "floor",
    "full",
    "full_like",
    "gather",
    "ge",
    "gt",
    "igamma",
    "igammac",
    "imag",
    "index_in_dim",
    "index_take",
    "integer_pow",
    "iota",
    "is_finite",
    "le",
    "lgamma",
    "log",
    "log1p",
    "logistic",
    "lt",
    "max",
    "min",
    "mul",
    "ne",
    "neg",
    "nextafter",
    "pad",
    "polygamma",
    "population_count",
    "pow",
    "random_gamma_grad",
    "real",
    "reciprocal",
    "reduce",
    "reduce_precision",
    "reduce_window",
    "rem",
    "reshape",
    "rev",
    "rng_bit_generator",
    "rng_uniform",
    "round",
    "rsqrt",
    "scatter",
    "scatter_add",
    "scatter_apply",
    "scatter_max",
    "scatter_min",
    "scatter_mul",
    "shift_left",
    "shift_right_arithmetic",
    "shift_right_logical",
    "sign",
    "sin",
    "sinh",
    "slice",
    "slice_in_dim",
    "sort",
    "sort_key_val",
    "sqrt",
    "square",
    "squeeze",
    "sub",
    "tan",
    "tanh",
    "top_k",
    "transpose",
    "zeros_like_array",
    "zeta",
    # ----- Control Flow Operators -----
    "associative_scan",
    "cond",
    "fori_loop",
    "map",
    "scan",
    "select",
    "select_n",
    "switch",
    "while_loop",
    # ----- Custom Gradient Operators -----
    "stop_gradient",
    "custom_linear_solve",
    "custom_root",
    # ----- Parallel Operators -----
    "all_gather",
    "all_to_all",
    "psum",
    "psum_scatter",
    "pmax",
    "pmin",
    "pmean",
    "ppermute",
    "pshuffle",
    "pswapaxes",
    "axis_index",
    # ----- Sharding-related Operators -----
    "with_sharding_constraint",
    # ----- Linear Algebra Operators -----
    "linalg",
]


import sys
from collections.abc import Callable
from typing import Any

from jax import lax
from quax import quaxify

from . import linalg


def __dir__() -> list[str]:
    """List the module contents."""
    return sorted(__all__)


# TODO: return type hint signature
def __getattr__(name: str) -> Callable[..., Any]:
    """Get the :external:`quax.quaxify`'ed function."""
    if name not in __all__:
        msg = f"Cannot get {name} from quaxed.lax."
        raise AttributeError(msg)

    # Quaxify the operator
    out = quaxify(getattr(lax, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out



__all__ = ["argmax", "argmin", "nonzero", "where"]


from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def argmax(
    x: ArrayLike,
    /,
    *,
    axis: int | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.argmax(x, axis=axis, keepdims=keepdims)


@quaxify
def argmin(
    x: ArrayLike,
    /,
    *,
    axis: int | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.argmin(x, axis=axis, keepdims=keepdims)


@quaxify
def nonzero(x: ArrayLike, /) -> tuple[Value, ...]:
    return array_api.nonzero(x)


@quaxify
def where(condition: ArrayLike, x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.where(condition, x1, x2)



"""Copyright (c) 2023 Nathaniel Starkman. All rights reserved.

quaxed: Pre-quaxed libraries for multiple dispatch over abstract array types in JAX
"""

# pylint: disable=redefined-builtin

from jax.experimental.array_api import __array_api_version__
from jaxtyping import install_import_hook

with install_import_hook("quaxed", None):
    from . import (
        _constants,
        _creation_functions,
        _data_type_functions,
        _dispatch,
        _elementwise_functions,
        _indexing_functions,
        _linear_algebra_functions,
        _manipulation_functions,
        _searching_functions,
        _set_functions,
        _sorting_functions,
        _statistical_functions,
        _utility_functions,
        fft,
        linalg,
    )
    from ._constants import *
    from ._creation_functions import *
    from ._data_type_functions import *
    from ._dispatch import *
    from ._elementwise_functions import *
    from ._indexing_functions import *
    from ._linear_algebra_functions import *
    from ._manipulation_functions import *
    from ._searching_functions import *
    from ._set_functions import *
    from ._sorting_functions import *
    from ._statistical_functions import *
    from ._utility_functions import *

__all__ = ["__array_api_version__", "fft", "linalg"]
__all__ += _constants.__all__
__all__ += _creation_functions.__all__
__all__ += _data_type_functions.__all__
__all__ += _elementwise_functions.__all__
__all__ += _indexing_functions.__all__
__all__ += _linear_algebra_functions.__all__
__all__ += _manipulation_functions.__all__
__all__ += _searching_functions.__all__
__all__ += _set_functions.__all__
__all__ += _sorting_functions.__all__
__all__ += _statistical_functions.__all__
__all__ += _utility_functions.__all__
__all__ += _dispatch.__all__



__all__ = [
    "abs",
    "acos",
    "acosh",
    "add",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "bitwise_and",
    "bitwise_left_shift",
    "bitwise_invert",
    "bitwise_or",
    "bitwise_right_shift",
    "bitwise_xor",
    "ceil",
    "conj",
    "cos",
    "cosh",
    "divide",
    "equal",
    "exp",
    "expm1",
    "floor",
    "floor_divide",
    "greater",
    "greater_equal",
    "imag",
    "isfinite",
    "isinf",
    "isnan",
    "less",
    "less_equal",
    "log",
    "log1p",
    "log2",
    "log10",
    "logaddexp",
    "logical_and",
    "logical_not",
    "logical_or",
    "logical_xor",
    "multiply",
    "negative",
    "not_equal",
    "positive",
    "pow",
    "real",
    "remainder",
    "round",
    "sign",
    "sin",
    "sinh",
    "square",
    "sqrt",
    "subtract",
    "tan",
    "tanh",
    "trunc",
]


from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def abs(x: ArrayLike, /) -> Value:
    return array_api.abs(x)


@quaxify
def acos(x: ArrayLike, /) -> Value:
    return array_api.acos(x)


@quaxify
def acosh(x: ArrayLike, /) -> Value:
    return array_api.acosh(x)


@quaxify
def add(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.add(x1, x2)


@quaxify
def asin(x: ArrayLike, /) -> Value:
    return array_api.asin(x)


@quaxify
def asinh(x: ArrayLike, /) -> Value:
    return array_api.asinh(x)


@quaxify
def atan(x: ArrayLike, /) -> Value:
    return array_api.atan(x)


@quaxify
def atan2(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.atan2(x1, x2)


@quaxify
def atanh(x: ArrayLike, /) -> Value:
    return array_api.atanh(x)


@quaxify
def bitwise_and(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.bitwise_and(x1, x2)


@quaxify
def bitwise_left_shift(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.bitwise_left_shift(x1, x2)


@quaxify
def bitwise_invert(x: ArrayLike, /) -> Value:
    return array_api.bitwise_invert(x)


@quaxify
def bitwise_or(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.bitwise_or(x1, x2)


@quaxify
def bitwise_right_shift(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.bitwise_right_shift(x1, x2)


@quaxify
def bitwise_xor(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.bitwise_xor(x1, x2)


@quaxify
def ceil(x: ArrayLike, /) -> Value:
    return array_api.ceil(x)


@quaxify
def conj(x: ArrayLike, /) -> Value:
    return array_api.conj(x)


@quaxify
def cos(x: ArrayLike, /) -> Value:
    return array_api.cos(x)


@quaxify
def cosh(x: ArrayLike, /) -> Value:
    return array_api.cosh(x)


@quaxify
def divide(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.divide(x1, x2)


@quaxify
def equal(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.equal(x1, x2)


@quaxify
def exp(x: ArrayLike, /) -> Value:
    return array_api.exp(x)


@quaxify
def expm1(x: ArrayLike, /) -> Value:
    return array_api.expm1(x)


@quaxify
def floor(x: ArrayLike, /) -> Value:
    return array_api.floor(x)


@quaxify
def floor_divide(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.floor_divide(x1, x2)


@quaxify
def greater(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.greater(x1, x2)


@quaxify
def greater_equal(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.greater_equal(x1, x2)


@quaxify
def imag(x: ArrayLike, /) -> Value:
    return array_api.imag(x)


@quaxify
def isfinite(x: ArrayLike, /) -> Value:
    return array_api.isfinite(x)


@quaxify
def isinf(x: ArrayLike, /) -> Value:
    # Jax `isinf` makes a numpy array with value `inf` and then compares it with
    # the input array. If the input array cannot be compared to base numpy
    # arrays, e.g. a Quantity with units, then Jax's `isinf` will raise an
    # unwanted error. Instead, we just negate the `isfinite` function, which
    # should work for all array-like inputs.
    return ~array_api.isfinite(x)


@quaxify
def isnan(x: ArrayLike, /) -> Value:
    return array_api.isnan(x)


@quaxify
def less(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.less(x1, x2)


@quaxify
def less_equal(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.less_equal(x1, x2)


@quaxify
def log(x: ArrayLike, /) -> Value:
    return array_api.log(x)


@quaxify
def log1p(x: ArrayLike, /) -> Value:
    return array_api.log1p(x)


@quaxify
def log2(x: ArrayLike, /) -> Value:
    return array_api.log2(x)


@quaxify
def log10(x: ArrayLike, /) -> Value:
    return array_api.log10(x)


@quaxify
def logaddexp(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.logaddexp(x1, x2)


@quaxify
def logical_and(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.logical_and(x1, x2)


@quaxify
def logical_not(x: ArrayLike, /) -> Value:
    return array_api.logical_not(x)


@quaxify
def logical_or(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.logical_or(x1, x2)


@quaxify
def logical_xor(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.logical_xor(x1, x2)


@quaxify
def multiply(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.multiply(x1, x2)


@quaxify
def negative(x: ArrayLike, /) -> Value:
    return array_api.negative(x)


@quaxify
def not_equal(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.not_equal(x1, x2)


@quaxify
def positive(x: ArrayLike, /) -> Value:
    return array_api.positive(x)


@quaxify
def pow(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.pow(x1, x2)


@quaxify
def real(x: ArrayLike, /) -> Value:
    return array_api.real(x)


@quaxify
def remainder(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.remainder(x1, x2)


@quaxify
def round(x: ArrayLike, /) -> Value:
    return array_api.round(x)


@quaxify
def sign(x: ArrayLike, /) -> Value:
    return array_api.sign(x)


@quaxify
def sin(x: ArrayLike, /) -> Value:
    return array_api.sin(x)


@quaxify
def sinh(x: ArrayLike, /) -> Value:
    return array_api.sinh(x)


@quaxify
def square(x: ArrayLike, /) -> Value:
    return array_api.square(x)


@quaxify
def sqrt(x: ArrayLike, /) -> Value:
    return array_api.sqrt(x)


@quaxify
def subtract(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.subtract(x1, x2)


@quaxify
def tan(x: ArrayLike, /) -> Value:
    return array_api.tan(x)


@quaxify
def tanh(x: ArrayLike, /) -> Value:
    return array_api.tanh(x)


@quaxify
def trunc(x: ArrayLike, /) -> Value:
    return array_api.trunc(x)



"""Test with :class:`MyArray` inputs."""

from collections.abc import Callable, Sequence
from dataclasses import replace
from typing import Any

import equinox as eqx
import jax
import jax.experimental.array_api as jax_xp
from jax import Device, lax
from jax._src.lax.slicing import GatherDimensionNumbers, GatherScatterMode
from jax._src.typing import Shape
from jaxtyping import ArrayLike
from plum import Dispatcher, Function, dispatch
from quax import ArrayValue, register

from quaxed._types import DType
from quaxed.array_api._dispatch import dispatcher as ap_dispatcher


class MyArray(ArrayValue):
    """A :class:`quax.ArrayValue` that is dense.

    This is different from :class:`quax.MyArray` only in that
    `quax` will not attempt to convert it to a JAX array.
    """

    array: jax.Array = eqx.field(converter=jax_xp.asarray)

    def materialise(self) -> jax.Array:
        """Convert to a JAX array."""
        raise NotImplementedError

    def aval(self) -> jax.core.ShapedArray:
        """Return the ShapedArray."""
        return jax.core.get_aval(self.array)


# ==============================================================================


def chain_dispatchers(*dispatchers: Dispatcher) -> Callable[[Any], Function]:
    """Apply many dispatchers to a function."""

    def decorator(method: Any) -> Function:
        for dispatcher in dispatchers:
            f = dispatcher(method)
        return f

    return decorator


# ==============================================================================


@register(lax.abs_p)
def _abs_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.abs(x.array))


# ==============================================================================


@register(lax.acos_p)
def _acos_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.acos(x.array))


# ==============================================================================


@register(lax.acosh_p)
def _acosh_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.acosh(x.array))


# ==============================================================================


@register(lax.add_p)
def _add_p_qq(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.add(x.array, y))


@register(lax.add_p)
def _add_p_qq(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.add(x.array, y.array))


# ==============================================================================


@register(lax.after_all_p)
def _after_all_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.all_gather_p)
def _all_gather_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.all_to_all_p)
def _all_to_all_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.and_p)
def _and_p(x1: MyArray, x2: MyArray, /) -> MyArray:
    return MyArray(x1.array & x2.array)


# ==============================================================================


@register(lax.approx_top_k_p)
def _approx_top_k_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.argmax_p)
def _argmax_p(operand: MyArray, *, axes: Any, index_dtype: Any) -> MyArray:
    return replace(operand, array=lax.argmax(operand.array, axes[0], index_dtype))


# ==============================================================================


@register(lax.argmin_p)
def _argmin_p(operand: MyArray, *, axes: Any, index_dtype: Any) -> MyArray:
    return replace(operand, array=lax.argmin(operand.array, axes[0], index_dtype))


# ==============================================================================


@register(lax.asin_p)
def _asin_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.asin(x.array))


# ==============================================================================


@register(lax.asinh_p)
def _asinh_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.asinh(x.array))


# ==============================================================================


@register(lax.atan2_p)
def _atan2_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.atan2(x.array, y.array))


# ==============================================================================


@register(lax.atan_p)
def _atan_p(x: MyArray) -> MyArray:
    return MyArray(lax.atan(x.array))


# ==============================================================================


@register(lax.atanh_p)
def _atanh_p(x: MyArray) -> MyArray:
    return MyArray(lax.atanh(x.array))


# ==============================================================================


@register(lax.axis_index_p)
def _axis_index_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.bessel_i0e_p)
def _bessel_i0e_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.bessel_i1e_p)
def _bessel_i1e_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.bitcast_convert_type_p)
def _bitcast_convert_type_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.broadcast_in_dim_p)
def _broadcast_in_dim_p(operand: MyArray, **kwargs: Any) -> MyArray:
    return replace(operand, array=lax.broadcast_in_dim(operand.array, **kwargs))


# ==============================================================================


@register(lax.cbrt_p)
def _cbrt_p(x: MyArray) -> MyArray:
    return MyArray(lax.cbrt(x.array))


# ==============================================================================


@register(lax.ceil_p)
def _ceil_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.ceil(x.array))


# ==============================================================================


@register(lax.clamp_p)
def _clamp_p(min: MyArray, x: MyArray, max: MyArray) -> MyArray:
    return replace(x, array=lax.clamp(min.array, x.array, max.array))


# ==============================================================================


@register(lax.clz_p)
def _clz_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.complex_p)
def _complex_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.complex(x.array, y.array))


# ==============================================================================


@register(lax.concatenate_p)
def _concatenate_p(
    operand0: MyArray,
    *operands: MyArray,
    **kwargs: Any,
) -> MyArray:
    return MyArray(
        lax.concatenate([operand0.array] + [op.array for op in operands], **kwargs)
    )


@register(lax.concatenate_p)
def _concatenate_p(operand0: ArrayLike, operand1: MyArray, /, **kwargs: Any) -> MyArray:
    return MyArray(lax.concatenate_p.bind(operand0, operand1.array, **kwargs))


# ==============================================================================


@register(lax.cond_p)  # TODO: implement
def _cond_p(index, consts) -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.conj_p)
def _conj_p(x: MyArray, *, input_dtype: Any) -> MyArray:
    del input_dtype  # TODO: use this?
    return replace(x, array=lax.conj(x.array))


# ==============================================================================


@register(lax.conv_general_dilated_p)
def _conv_general_dilated_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.convert_element_type_p)
def _convert_element_type_p(operand: MyArray, **kwargs: Any) -> MyArray:
    return replace(
        operand,
        array=lax.convert_element_type_p.bind(operand.array, **kwargs),
    )


# ==============================================================================


@register(lax.copy_p)
def _copy_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.copy_p.bind(x.array))


# ==============================================================================


@register(lax.cos_p)
def _cos_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.cos(x.array))


# ==============================================================================


@register(lax.cosh_p)
def _cosh_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.cosh(x.array))


# ==============================================================================


@register(lax.create_token_p)
def _create_token_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.cumlogsumexp_p)
def _cumlogsumexp_p(operand: MyArray, *, axis: Any, reverse: Any) -> MyArray:
    # TODO: double check units make sense here.
    return replace(
        operand,
        array=lax.cumlogsumexp(operand.array, axis=axis, reverse=reverse),
    )


# ==============================================================================


@register(lax.cummax_p)
def _cummax_p(operand: MyArray, *, axis: Any, reverse: Any) -> MyArray:
    return replace(operand, array=lax.cummax(operand.array, axis=axis, reverse=reverse))


# ==============================================================================


@register(lax.cummin_p)
def _cummin_p(operand: MyArray, *, axis: Any, reverse: Any) -> MyArray:
    return replace(operand, array=lax.cummin(operand.array, axis=axis, reverse=reverse))


# ==============================================================================


@register(lax.cumprod_p)
def _cumprod_p(operand: MyArray, *, axis: Any, reverse: Any) -> MyArray:
    return replace(
        operand,
        array=lax.cumprod(operand.array, axis=axis, reverse=reverse),
    )


# ==============================================================================


@register(lax.cumsum_p)
def _cumsum_p(operand: MyArray, *, axis: Any, reverse: Any) -> MyArray:
    return replace(operand, array=lax.cumsum(operand.array, axis=axis, reverse=reverse))


# ==============================================================================


@register(lax.device_put_p)
def _device_put_p(x: MyArray, **kwargs: Any) -> MyArray:
    return replace(x, array=jax.device_put(x.array, **kwargs))


# ==============================================================================


@register(lax.digamma_p)
def _digamma_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.digamma(x.array))


# ==============================================================================


@register(lax.div_p)
def _div_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.div(x.array, y.array))


@register(lax.div_p)
def _div_p(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.div(x.array, y))


# ==============================================================================


@register(lax.dot_general_p)  # TODO: implement
def _dot_general_p(lhs: MyArray, rhs: MyArray, **kwargs: Any) -> MyArray:
    return MyArray(lax.dot_general_p.bind(lhs.array, rhs.array, **kwargs))


# ==============================================================================


@register(lax.dynamic_slice_p)
def _dynamic_slice_p(
    operand: MyArray,
    start_indices: ArrayLike,
    dynamic_sizes: ArrayLike,
    *,
    slice_sizes: Any,
) -> MyArray:
    raise NotImplementedError  # TODO: implement


# ==============================================================================


@register(lax.dynamic_update_slice_p)
def _dynamic_update_slice_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.eq_p)
def _eq_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.eq(x.array, y.array))


@register(lax.eq_p)
def _eq_p(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.eq(x.array, y))


# ==============================================================================


@register(lax.eq_to_p)
def _eq_to_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.erf_inv_p)
def _erf_inv_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.erf_inv(x.array))


# ==============================================================================


@register(lax.erf_p)
def _erf_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.erf(x.array))


# ==============================================================================


@register(lax.erfc_p)
def _erfc_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.erfc(x.array))


# ==============================================================================


@register(lax.exp2_p)
def _exp2_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.exp2(x.array))


# ==============================================================================


@register(lax.exp_p)
def _exp_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.exp(x.array))


# ==============================================================================


@register(lax.expm1_p)
def _expm1_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.expm1(x.array))


# ==============================================================================


@register(lax.fft_p)
def _fft_p(x: MyArray, *, fft_type: Any, fft_lengths: Any) -> MyArray:
    return replace(x, array=lax.fft(x.array, fft_type, fft_lengths))


# ==============================================================================


@register(lax.floor_p)
def _floor_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.floor(x.array))


# ==============================================================================


@register(lax.gather_p)
def _gather_p(
    operand: MyArray,
    start_indices: MyArray,
    *,
    dimension_numbers: GatherDimensionNumbers,
    slice_sizes: Shape,
    unique_indices: bool,
    indices_are_sorted: bool,
    mode: str | GatherScatterMode | None = None,
    fill_value: Any = None,
) -> MyArray:
    return MyArray(
        lax.gather(
            operand.array,
            start_indices.array,
            dimension_numbers=dimension_numbers,
            slice_sizes=slice_sizes,
            unique_indices=unique_indices,
            indices_are_sorted=indices_are_sorted,
            mode=mode,
            fill_value=fill_value,
        ),
    )


@register(lax.gather_p)
def _gather_p(
    operand: MyArray,
    start_indices: ArrayLike,
    *,
    dimension_numbers: GatherDimensionNumbers,
    slice_sizes: Shape,
    unique_indices: bool,
    indices_are_sorted: bool,
    mode: str | GatherScatterMode | None = None,
    fill_value: Any = None,
) -> MyArray:
    return MyArray(
        lax.gather(
            operand.array,
            start_indices,
            dimension_numbers=dimension_numbers,
            slice_sizes=slice_sizes,
            unique_indices=unique_indices,
            indices_are_sorted=indices_are_sorted,
            mode=mode,
            fill_value=fill_value,
        ),
    )


# ==============================================================================


@register(lax.ge_p)
def _ge_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.ge(x.array, y.array))


# ==============================================================================


@register(lax.gt_p)
def _gt_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.gt(x.array, y.array))


# ==============================================================================


@register(lax.igamma_grad_a_p)
def _igamma_grad_a_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.igamma_p)
def _igamma_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.igammac_p)
def _igammac_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.imag_p)
def _imag_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.imag(x.array))


# ==============================================================================


@register(lax.infeed_p)
def _infeed_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.integer_pow_p)
def _integer_pow_p(x: MyArray, *, y: Any) -> MyArray:
    return replace(x, array=lax.integer_pow(x.array, y))


# ==============================================================================


# @register(lax.iota_p)
# def _iota_p(dtype: MyArray) -> MyArray:
#     raise NotImplementedError


# ==============================================================================


@register(lax.is_finite_p)
def _is_finite_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.is_finite(x.array))


# ==============================================================================


@register(lax.le_p)
def _le_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.le(x.array, y.array))


# ==============================================================================


@register(lax.le_to_p)
def _le_to_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.lgamma_p)
def _lgamma_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.lgamma(x.array))


# ==============================================================================


@register(lax.linear_solve_p)
def _linear_solve_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.log1p_p)
def _log1p_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.log1p(x.array))


# ==============================================================================


@register(lax.log_p)
def _log_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.log(x.array))


# ==============================================================================


@register(lax.logistic_p)
def _logistic_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.logistic(x.array))


# ==============================================================================


@register(lax.lt_p)
def _lt_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.lt(x.array, y.array))


@register(lax.lt_p)
def _lt_p(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.lt(x.array, y))


# ==============================================================================


@register(lax.lt_to_p)
def _lt_to_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.max_p)
def _max_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.max(x.array, y.array))


@register(lax.max_p)
def _max_p_d1(x: ArrayLike, y: MyArray) -> MyArray:
    return MyArray(lax.max(x, y.array))


@register(lax.max_p)
def _max_p_d2(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.max(x.array, y))


# ==============================================================================


@register(lax.min_p)
def _min_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.min(x.array, y.array))


# ==============================================================================
# Multiplication


@register(lax.mul_p)
def _mul_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.mul(x.array, y.array))


# ==============================================================================


@register(lax.ne_p)
def _ne_p(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.ne(x.array, y))


@register(lax.ne_p)
def _ne_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.ne(x.array, y.array))


# ==============================================================================


@register(lax.neg_p)
def _neg_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.neg(x.array))


# ==============================================================================


@register(lax.nextafter_p)
def _nextafter_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.not_p)
def _not_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.bitwise_not(x.array))


# ==============================================================================


@register(lax.or_p)
def _or_p(x: MyArray, y: MyArray) -> MyArray:
    return replace(x, array=lax.bitwise_or(x.array, y.array))


# ==============================================================================


@register(lax.outfeed_p)
def _outfeed_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.pad_p)
def _pad_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.pmax_p)
def _pmax_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.pmin_p)
def _pmin_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.polygamma_p)
def _polygamma_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.population_count_p)
def _population_count_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.pow_p)
def _pow_p_qq(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(array=lax.pow(x.array, y.array))


# ==============================================================================


@register(lax.ppermute_p)
def _ppermute_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.psum_p)
def _psum_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.random_gamma_grad_p)
def _random_gamma_grad_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.real_p)
def _real_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.real(x.array))


# ==============================================================================


@register(lax.reduce_and_p)
def _reduce_and_p(
    operand: MyArray,
    *,
    axes: Sequence[int],
) -> Any:
    return lax.reduce_and_p.bind(operand.array, axes=tuple(axes))


# ==============================================================================


@register(lax.reduce_max_p)
def _reduce_max_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_min_p)
def _reduce_min_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_or_p)
def _reduce_or_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_p)
def _reduce_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_precision_p)
def _reduce_precision_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_prod_p)
def _reduce_prod_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_sum_p)
def _reduce_sum_p(operand: MyArray, *, axes: tuple[int, ...]) -> MyArray:
    return MyArray(lax.reduce_sum_p.bind(operand.array, axes=axes))


# ==============================================================================


@register(lax.reduce_window_max_p)
def _reduce_window_max_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_window_min_p)
def _reduce_window_min_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_window_p)
def _reduce_window_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_window_sum_p)
def _reduce_window_sum_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.reduce_xor_p)
def _reduce_xor_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.regularized_incomplete_beta_p)
def _regularized_incomplete_beta_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.rem_p)
def _rem_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.rem(x.array, y.array))


@register(lax.rem_p)
def _rem_p_d1(x: ArrayLike, y: MyArray) -> MyArray:
    return MyArray(lax.rem(x, y.array))


@register(lax.rem_p)
def _rem_p_d1(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.rem(x.array, y))


# ==============================================================================


@register(lax.reshape_p)
def _reshape_p(operand: MyArray, *, new_sizes: Any, dimensions: Any) -> MyArray:
    return replace(operand, array=lax.reshape(operand.array, new_sizes, dimensions))


# ==============================================================================


@register(lax.rev_p)
def _rev_p(operand: MyArray, *, dimensions: Any) -> MyArray:
    return replace(operand, array=lax.rev(operand.array, dimensions))


# ==============================================================================


@register(lax.rng_bit_generator_p)
def _rng_bit_generator_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.rng_uniform_p)
def _rng_uniform_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.round_p)
def _round_p(x: MyArray, *, rounding_method: Any) -> MyArray:
    return replace(x, array=lax.round(x.array, rounding_method))


# ==============================================================================


@register(lax.rsqrt_p)
def _rsqrt_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.rsqrt(x.array) ** (-1 / 2))


# ==============================================================================


@register(lax.scan_p)
def _scan_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.scatter_add_p)
def _scatter_add_p(
    operand: MyArray,
    scatter_indices: MyArray,
    updates: MyArray,
    *,
    update_jaxpr: Any,
    update_consts: Any,
    dimension_numbers: Any,
    indices_are_sorted: bool,
    unique_indices: bool,
    mode: str | GatherScatterMode | None = None,
) -> MyArray:
    return MyArray(
        lax.scatter_add_p.bind(
            operand.array,
            scatter_indices.array,
            updates.array,
            update_jaxpr=update_jaxpr,
            update_consts=update_consts,
            dimension_numbers=dimension_numbers,
            indices_are_sorted=indices_are_sorted,
            unique_indices=unique_indices,
            mode=mode,
        ),
    )


@register(lax.scatter_add_p)
def _scatter_add_p(
    operand: MyArray,
    scatter_indices: ArrayLike,
    updates: ArrayLike,
    *,
    update_jaxpr: Any,
    update_consts: Any,
    dimension_numbers: Any,
    indices_are_sorted: bool,
    unique_indices: bool,
    mode: str | GatherScatterMode | None = None,
) -> MyArray:
    return MyArray(
        lax.scatter_add_p.bind(
            operand.array,
            scatter_indices,
            updates,
            update_jaxpr=update_jaxpr,
            update_consts=update_consts,
            dimension_numbers=dimension_numbers,
            indices_are_sorted=indices_are_sorted,
            unique_indices=unique_indices,
            mode=mode,
        ),
    )


@register(lax.scatter_add_p)
def _scatter_add_p(
    operand: ArrayLike,
    scatter_indices: MyArray,
    updates: ArrayLike,
    *,
    update_jaxpr: Any,
    update_consts: Any,
    dimension_numbers: Any,
    indices_are_sorted: bool,
    unique_indices: bool,
    mode: str | GatherScatterMode | None = None,
) -> MyArray:
    return MyArray(
        lax.scatter_add_p.bind(
            operand,
            scatter_indices.array,
            updates,
            update_jaxpr=update_jaxpr,
            update_consts=update_consts,
            dimension_numbers=dimension_numbers,
            indices_are_sorted=indices_are_sorted,
            unique_indices=unique_indices,
            mode=mode,
        ),
    )


# @register(lax.scatter_add_p)
# def _scatter_add_p(
#     operand: Zero,
#     scatter_indices: MyArray,
#     updates: MyArray,
#     *,
#     update_jaxpr: Any,
#     update_consts: Any,
#     dimension_numbers: Any,
#     indices_are_sorted: bool,
#     unique_indices: bool,
#     mode: str | GatherScatterMode | None = None,
# ) -> MyArray:
#     return MyArray(
#         lax.scatter_add_p.bind(
#             jax_xp.zeros_like(operand),
#             scatter_indices.array,
#             updates.array,
#             update_jaxpr=update_jaxpr,
#             update_consts=update_consts,
#             dimension_numbers=dimension_numbers,
#             indices_are_sorted=indices_are_sorted,
#             unique_indices=unique_indices,
#             mode=mode,
#         ),
#     )


# ==============================================================================


@register(lax.scatter_max_p)
def _scatter_max_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.scatter_min_p)
def _scatter_min_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.scatter_mul_p)
def _scatter_mul_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.scatter_p)
def _scatter_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.select_and_gather_add_p)
def _select_and_gather_add_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.select_and_scatter_add_p)
def _select_and_scatter_add_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.select_and_scatter_p)
def _select_and_scatter_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.select_n_p)
def _select_n_p(which: MyArray, *cases: MyArray) -> MyArray:
    # Process the cases, replacing Zero and MyArray with a materialised array.
    return MyArray(lax.select_n(which.array, *(case.array for case in cases)))


@register(lax.select_n_p)
def _select_n_p(which: ArrayLike, case0: ArrayLike, case1: MyArray) -> MyArray:
    return MyArray(lax.select_n(which, case0, case1.array))


@register(lax.select_n_p)
def _select_n_p(which: ArrayLike, case0: MyArray, case1: ArrayLike) -> MyArray:
    return MyArray(lax.select_n(which, case0.array, case1))


# ==============================================================================


@register(lax.sharding_constraint_p)
def _sharding_constraint_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.shift_left_p)
def _shift_left_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.shift_left(x.array, y.array))


# ==============================================================================


@register(lax.shift_right_arithmetic_p)
def _shift_right_arithmetic_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.shift_right_arithmetic(x.array, y.array))


# ==============================================================================


@register(lax.shift_right_logical_p)
def _shift_right_logical_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.sign_p)
def _sign_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.sign(x.array))


# ==============================================================================


@register(lax.sin_p)
def _sin_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.sin(x.array))


# ==============================================================================


@register(lax.sinh_p)
def _sinh_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.sinh(x.array))


# ==============================================================================


@register(lax.slice_p)
def _slice_p(
    operand: MyArray,
    *,
    start_indices: Any,
    limit_indices: Any,
    strides: Any,
) -> MyArray:
    return replace(
        operand,
        array=lax.slice_p.bind(
            operand.array,
            start_indices=start_indices,
            limit_indices=limit_indices,
            strides=strides,
        ),
    )


# ==============================================================================


@register(lax.sort_p)
def _sort_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.sqrt_p)
def _sqrt_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.sqrt(x.array))


# ==============================================================================


@register(lax.squeeze_p)
def _squeeze_p(x: MyArray, *, dimensions: Any) -> MyArray:
    return replace(x, array=lax.squeeze(x.array, dimensions))


# ==============================================================================


@register(lax.stop_gradient_p)
def _stop_gradient_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.stop_gradient(x.array))


# ==============================================================================
# Subtraction


@register(lax.sub_p)
def _sub_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.sub(x.array, y.array))


@register(lax.sub_p)
def _sub_p(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.sub(x.array, y))


# ==============================================================================


@register(lax.tan_p)
def _tan_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.tan(x.array))


# ==============================================================================


@register(lax.tanh_p)
def _tanh_p(x: MyArray) -> MyArray:
    return replace(x, array=lax.tanh(x.array))


# ==============================================================================


@register(lax.top_k_p)
def _top_k_p(operand: MyArray, k: int) -> MyArray:
    raise replace(operand, array=lax.top_k(operand.array, k))


# ==============================================================================


@register(lax.transpose_p)
def _transpose_p(operand: MyArray, *, permutation: Any) -> MyArray:
    return replace(operand, array=lax.transpose(operand.array, permutation))


# ==============================================================================


@register(lax.while_p)
def _while_p() -> MyArray:
    raise NotImplementedError


# ==============================================================================


@register(lax.xor_p)
def _xor_p(x: MyArray, y: MyArray) -> MyArray:
    return MyArray(lax.bitwise_xor(x.array, y.array))


@register(lax.xor_p)
def _xor_p(x: MyArray, y: ArrayLike) -> MyArray:
    return MyArray(lax.bitwise_xor(x.array, y))


# ==============================================================================


@register(lax.zeta_p)
def _zeta_p() -> MyArray:
    raise NotImplementedError


###############################################################################


@chain_dispatchers(dispatch, ap_dispatcher)
def arange(
    start: MyArray,
    stop: MyArray | None = None,
    step: MyArray | None = None,
    *,
    dtype: Any = None,
    device: Any = None,
) -> MyArray:
    return MyArray(
        jax_xp.arange(
            start.array,
            stop=stop.array if stop is not None else None,
            step=step.array if step is not None else None,
            dtype=dtype,
            device=device,
        ),
    )


@chain_dispatchers(dispatch, ap_dispatcher)
def empty_like(
    x: MyArray,
    /,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> MyArray:
    return MyArray(jax_xp.empty_like(x.array, dtype=dtype, device=device))


@chain_dispatchers(dispatch, ap_dispatcher)
def full_like(
    x: MyArray,
    /,
    fill_value: bool | int | float | complex | MyArray,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> MyArray:
    return MyArray(
        jax_xp.full_like(x.array, fill_value, dtype=dtype, device=device),
    )


@chain_dispatchers(dispatch, ap_dispatcher)
def linspace(
    start: MyArray,
    stop: MyArray,
    num: int,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
    endpoint: bool = True,
) -> MyArray:
    return MyArray(
        jax_xp.linspace(
            start.array,
            stop.array,
            num=num,
            dtype=dtype,
            device=device,
            endpoint=endpoint,
        ),
    )


@chain_dispatchers(dispatch, ap_dispatcher)
def ones_like(
    x: MyArray,
    /,
    dtype: DType | None = None,
    device: Device | None = None,
) -> MyArray:
    return MyArray(jax_xp.ones_like(x.array, dtype=dtype, device=device))


@chain_dispatchers(dispatch, ap_dispatcher)
def zeros_like(
    x: MyArray,
    /,
    dtype: DType | None = None,
    device: Device | None = None,
) -> MyArray:
    return MyArray(jax_xp.zeros_like(x.array, dtype=dtype, device=device))



# Modules
from jax.numpy import fft as fft
from jax.numpy import linalg as linalg

# _core.py
# isort: split
from jax.numpy import abs as abs
from jax.numpy import absolute as absolute
from jax.numpy import acos as acos
from jax.numpy import acosh as acosh
from jax.numpy import add as add
from jax.numpy import all as all
from jax.numpy import allclose as allclose
from jax.numpy import amax as amax
from jax.numpy import amin as amin
from jax.numpy import angle as angle
from jax.numpy import any as any
from jax.numpy import append as append
from jax.numpy import apply_along_axis as apply_along_axis
from jax.numpy import apply_over_axes as apply_over_axes
from jax.numpy import arccos as arccos
from jax.numpy import arccosh as arccosh
from jax.numpy import arcsin as arcsin
from jax.numpy import arcsinh as arcsinh
from jax.numpy import arctan as arctan
from jax.numpy import arctan2 as arctan2
from jax.numpy import arctanh as arctanh
from jax.numpy import argmax as argmax
from jax.numpy import argmin as argmin
from jax.numpy import argpartition as argpartition
from jax.numpy import argsort as argsort
from jax.numpy import argwhere as argwhere
from jax.numpy import around as around
from jax.numpy import array as array
from jax.numpy import array_equal as array_equal
from jax.numpy import array_equiv as array_equiv
from jax.numpy import array_repr as array_repr
from jax.numpy import array_split as array_split
from jax.numpy import asin as asin
from jax.numpy import asinh as asinh
from jax.numpy import astype as astype
from jax.numpy import atan as atan
from jax.numpy import atan2 as atan2
from jax.numpy import atanh as atanh
from jax.numpy import atleast_1d as atleast_1d
from jax.numpy import atleast_2d as atleast_2d
from jax.numpy import atleast_3d as atleast_3d
from jax.numpy import average as average
from jax.numpy import bartlett as bartlett
from jax.numpy import bfloat16 as bfloat16
from jax.numpy import bincount as bincount
from jax.numpy import bitwise_and as bitwise_and
from jax.numpy import bitwise_count as bitwise_count
from jax.numpy import bitwise_invert as bitwise_invert
from jax.numpy import bitwise_left_shift as bitwise_left_shift
from jax.numpy import bitwise_not as bitwise_not
from jax.numpy import bitwise_or as bitwise_or
from jax.numpy import bitwise_right_shift as bitwise_right_shift
from jax.numpy import bitwise_xor as bitwise_xor
from jax.numpy import blackman as blackman
from jax.numpy import block as block
from jax.numpy import bool as bool
from jax.numpy import bool_ as bool_
from jax.numpy import broadcast_arrays as broadcast_arrays
from jax.numpy import broadcast_shapes as broadcast_shapes
from jax.numpy import broadcast_to as broadcast_to
from jax.numpy import c_ as c_
from jax.numpy import can_cast as can_cast
from jax.numpy import cbrt as cbrt
from jax.numpy import cdouble as cdouble
from jax.numpy import ceil as ceil
from jax.numpy import character as character
from jax.numpy import choose as choose
from jax.numpy import clip as clip
from jax.numpy import column_stack as column_stack
from jax.numpy import complex64 as complex64
from jax.numpy import complex128 as complex128
from jax.numpy import complex_ as complex_
from jax.numpy import complexfloating as complexfloating
from jax.numpy import compress as compress
from jax.numpy import concat as concat
from jax.numpy import concatenate as concatenate
from jax.numpy import conj as conj
from jax.numpy import conjugate as conjugate
from jax.numpy import convolve as convolve
from jax.numpy import copy as copy
from jax.numpy import copysign as copysign
from jax.numpy import corrcoef as corrcoef
from jax.numpy import correlate as correlate
from jax.numpy import cos as cos
from jax.numpy import cosh as cosh
from jax.numpy import count_nonzero as count_nonzero
from jax.numpy import cov as cov
from jax.numpy import cross as cross
from jax.numpy import csingle as csingle
from jax.numpy import cumprod as cumprod
from jax.numpy import cumsum as cumsum
from jax.numpy import deg2rad as deg2rad
from jax.numpy import degrees as degrees
from jax.numpy import delete as delete
from jax.numpy import diag as diag
from jax.numpy import diag_indices as diag_indices
from jax.numpy import diag_indices_from as diag_indices_from
from jax.numpy import diagflat as diagflat
from jax.numpy import diagonal as diagonal
from jax.numpy import diff as diff
from jax.numpy import digitize as digitize
from jax.numpy import divide as divide
from jax.numpy import divmod as divmod
from jax.numpy import dot as dot
from jax.numpy import double as double
from jax.numpy import dsplit as dsplit
from jax.numpy import dstack as dstack
from jax.numpy import dtype as dtype
from jax.numpy import e as e
from jax.numpy import ediff1d as ediff1d
from jax.numpy import einsum as einsum
from jax.numpy import einsum_path as einsum_path
from jax.numpy import empty as empty
from jax.numpy import equal as equal
from jax.numpy import euler_gamma as euler_gamma
from jax.numpy import exp as exp
from jax.numpy import exp2 as exp2
from jax.numpy import expand_dims as expand_dims
from jax.numpy import expm1 as expm1
from jax.numpy import extract as extract
from jax.numpy import eye as eye
from jax.numpy import fabs as fabs
from jax.numpy import fill_diagonal as fill_diagonal
from jax.numpy import finfo as finfo
from jax.numpy import fix as fix
from jax.numpy import flatnonzero as flatnonzero
from jax.numpy import flexible as flexible
from jax.numpy import flip as flip
from jax.numpy import fliplr as fliplr
from jax.numpy import flipud as flipud
from jax.numpy import float8_e4m3b11fnuz as float8_e4m3b11fnuz
from jax.numpy import float8_e4m3fn as float8_e4m3fn
from jax.numpy import float8_e4m3fnuz as float8_e4m3fnuz
from jax.numpy import float8_e5m2 as float8_e5m2
from jax.numpy import float8_e5m2fnuz as float8_e5m2fnuz
from jax.numpy import float16 as float16
from jax.numpy import float32 as float32
from jax.numpy import float64 as float64
from jax.numpy import float_ as float_
from jax.numpy import float_power as float_power
from jax.numpy import floating as floating
from jax.numpy import floor as floor
from jax.numpy import floor_divide as floor_divide
from jax.numpy import fmax as fmax
from jax.numpy import fmin as fmin
from jax.numpy import fmod as fmod
from jax.numpy import frexp as frexp
from jax.numpy import from_dlpack as from_dlpack
from jax.numpy import frombuffer as frombuffer
from jax.numpy import fromfunction as fromfunction
from jax.numpy import fromiter as fromiter
from jax.numpy import frompyfunc as frompyfunc
from jax.numpy import fromstring as fromstring
from jax.numpy import gcd as gcd
from jax.numpy import generic as generic
from jax.numpy import geomspace as geomspace
from jax.numpy import get_printoptions as get_printoptions
from jax.numpy import gradient as gradient
from jax.numpy import greater as greater
from jax.numpy import greater_equal as greater_equal
from jax.numpy import hamming as hamming
from jax.numpy import hanning as hanning
from jax.numpy import heaviside as heaviside
from jax.numpy import histogram as histogram
from jax.numpy import histogram2d as histogram2d
from jax.numpy import histogram_bin_edges as histogram_bin_edges
from jax.numpy import histogramdd as histogramdd
from jax.numpy import hsplit as hsplit
from jax.numpy import hstack as hstack
from jax.numpy import hypot as hypot
from jax.numpy import i0 as i0
from jax.numpy import identity as identity
from jax.numpy import iinfo as iinfo
from jax.numpy import imag as imag
from jax.numpy import index_exp as index_exp
from jax.numpy import indices as indices
from jax.numpy import inexact as inexact
from jax.numpy import inf as inf
from jax.numpy import inner as inner
from jax.numpy import insert as insert
from jax.numpy import int4 as int4
from jax.numpy import int8 as int8
from jax.numpy import int16 as int16
from jax.numpy import int32 as int32
from jax.numpy import int64 as int64
from jax.numpy import int_ as int_
from jax.numpy import integer as integer
from jax.numpy import interp as interp
from jax.numpy import intersect1d as intersect1d
from jax.numpy import invert as invert
from jax.numpy import isclose as isclose
from jax.numpy import iscomplex as iscomplex
from jax.numpy import iscomplexobj as iscomplexobj
from jax.numpy import isdtype as isdtype
from jax.numpy import isfinite as isfinite
from jax.numpy import isin as isin
from jax.numpy import isinf as isinf
from jax.numpy import isnan as isnan
from jax.numpy import isneginf as isneginf
from jax.numpy import isposinf as isposinf
from jax.numpy import isreal as isreal
from jax.numpy import isrealobj as isrealobj
from jax.numpy import isscalar as isscalar
from jax.numpy import issubdtype as issubdtype
from jax.numpy import iterable as iterable
from jax.numpy import ix_ as ix_
from jax.numpy import kaiser as kaiser
from jax.numpy import kron as kron
from jax.numpy import lcm as lcm
from jax.numpy import ldexp as ldexp
from jax.numpy import left_shift as left_shift
from jax.numpy import less as less
from jax.numpy import less_equal as less_equal
from jax.numpy import lexsort as lexsort
from jax.numpy import load as load
from jax.numpy import log as log
from jax.numpy import log1p as log1p
from jax.numpy import log2 as log2
from jax.numpy import log10 as log10
from jax.numpy import logaddexp as logaddexp
from jax.numpy import logaddexp2 as logaddexp2
from jax.numpy import logical_and as logical_and
from jax.numpy import logical_not as logical_not
from jax.numpy import logical_or as logical_or
from jax.numpy import logical_xor as logical_xor
from jax.numpy import logspace as logspace
from jax.numpy import mask_indices as mask_indices
from jax.numpy import matmul as matmul
from jax.numpy import matrix_transpose as matrix_transpose
from jax.numpy import max as max
from jax.numpy import maximum as maximum
from jax.numpy import mean as mean
from jax.numpy import median as median
from jax.numpy import mgrid as mgrid
from jax.numpy import min as min
from jax.numpy import minimum as minimum
from jax.numpy import mod as mod
from jax.numpy import modf as modf
from jax.numpy import moveaxis as moveaxis
from jax.numpy import multiply as multiply
from jax.numpy import nan as nan
from jax.numpy import nan_to_num as nan_to_num
from jax.numpy import nanargmax as nanargmax
from jax.numpy import nanargmin as nanargmin
from jax.numpy import nancumprod as nancumprod
from jax.numpy import nancumsum as nancumsum
from jax.numpy import nanmax as nanmax
from jax.numpy import nanmean as nanmean
from jax.numpy import nanmedian as nanmedian
from jax.numpy import nanmin as nanmin
from jax.numpy import nanpercentile as nanpercentile
from jax.numpy import nanprod as nanprod
from jax.numpy import nanquantile as nanquantile
from jax.numpy import nanstd as nanstd
from jax.numpy import nansum as nansum
from jax.numpy import nanvar as nanvar
from jax.numpy import ndarray as ndarray
from jax.numpy import ndim as ndim
from jax.numpy import negative as negative
from jax.numpy import newaxis as newaxis
from jax.numpy import nextafter as nextafter
from jax.numpy import nonzero as nonzero
from jax.numpy import not_equal as not_equal
from jax.numpy import number as number
from jax.numpy import object_ as object_
from jax.numpy import ogrid as ogrid
from jax.numpy import ones as ones
from jax.numpy import outer as outer
from jax.numpy import packbits as packbits
from jax.numpy import pad as pad
from jax.numpy import partition as partition
from jax.numpy import percentile as percentile
from jax.numpy import permute_dims as permute_dims
from jax.numpy import pi as pi
from jax.numpy import piecewise as piecewise
from jax.numpy import place as place
from jax.numpy import poly as poly
from jax.numpy import polyadd as polyadd
from jax.numpy import polyder as polyder
from jax.numpy import polydiv as polydiv
from jax.numpy import polyfit as polyfit
from jax.numpy import polyint as polyint
from jax.numpy import polymul as polymul
from jax.numpy import polysub as polysub
from jax.numpy import polyval as polyval
from jax.numpy import positive as positive
from jax.numpy import pow as pow
from jax.numpy import power as power
from jax.numpy import printoptions as printoptions
from jax.numpy import prod as prod
from jax.numpy import promote_types as promote_types
from jax.numpy import ptp as ptp
from jax.numpy import put as put
from jax.numpy import quantile as quantile
from jax.numpy import r_ as r_
from jax.numpy import rad2deg as rad2deg
from jax.numpy import radians as radians
from jax.numpy import ravel as ravel
from jax.numpy import ravel_multi_index as ravel_multi_index
from jax.numpy import real as real
from jax.numpy import reciprocal as reciprocal
from jax.numpy import remainder as remainder
from jax.numpy import repeat as repeat
from jax.numpy import reshape as reshape
from jax.numpy import resize as resize
from jax.numpy import result_type as result_type
from jax.numpy import right_shift as right_shift
from jax.numpy import rint as rint
from jax.numpy import roll as roll
from jax.numpy import rollaxis as rollaxis
from jax.numpy import roots as roots
from jax.numpy import rot90 as rot90
from jax.numpy import round as round
from jax.numpy import round_ as round_
from jax.numpy import s_ as s_
from jax.numpy import save as save
from jax.numpy import savez as savez
from jax.numpy import searchsorted as searchsorted
from jax.numpy import select as select
from jax.numpy import set_printoptions as set_printoptions
from jax.numpy import setdiff1d as setdiff1d
from jax.numpy import setxor1d as setxor1d
from jax.numpy import shape as shape
from jax.numpy import sign as sign
from jax.numpy import signbit as signbit
from jax.numpy import signedinteger as signedinteger
from jax.numpy import sin as sin
from jax.numpy import sinc as sinc
from jax.numpy import single as single
from jax.numpy import sinh as sinh
from jax.numpy import size as size
from jax.numpy import sort as sort
from jax.numpy import sort_complex as sort_complex
from jax.numpy import split as split
from jax.numpy import sqrt as sqrt
from jax.numpy import square as square
from jax.numpy import squeeze as squeeze
from jax.numpy import stack as stack
from jax.numpy import std as std
from jax.numpy import subtract as subtract
from jax.numpy import sum as sum
from jax.numpy import swapaxes as swapaxes
from jax.numpy import take as take
from jax.numpy import take_along_axis as take_along_axis
from jax.numpy import tan as tan
from jax.numpy import tanh as tanh
from jax.numpy import tensordot as tensordot
from jax.numpy import tile as tile
from jax.numpy import trace as trace
from jax.numpy import transpose as transpose
from jax.numpy import tri as tri
from jax.numpy import tril_indices as tril_indices
from jax.numpy import tril_indices_from as tril_indices_from
from jax.numpy import trim_zeros as trim_zeros
from jax.numpy import triu_indices as triu_indices
from jax.numpy import triu_indices_from as triu_indices_from
from jax.numpy import true_divide as true_divide
from jax.numpy import trunc as trunc
from jax.numpy import uint as uint
from jax.numpy import uint4 as uint4
from jax.numpy import uint8 as uint8
from jax.numpy import uint16 as uint16
from jax.numpy import uint32 as uint32
from jax.numpy import uint64 as uint64
from jax.numpy import union1d as union1d
from jax.numpy import unique as unique
from jax.numpy import unique_all as unique_all
from jax.numpy import unique_counts as unique_counts
from jax.numpy import unique_inverse as unique_inverse
from jax.numpy import unique_values as unique_values
from jax.numpy import unpackbits as unpackbits
from jax.numpy import unravel_index as unravel_index
from jax.numpy import unsignedinteger as unsignedinteger
from jax.numpy import unwrap as unwrap
from jax.numpy import vander as vander
from jax.numpy import var as var
from jax.numpy import vdot as vdot
from jax.numpy import vecdot as vecdot
from jax.numpy import vsplit as vsplit
from jax.numpy import vstack as vstack
from jax.numpy import where as where
from jax.numpy import zeros as zeros

# _creation_functions.py
# isort: split
from jax.numpy import arange as arange
from jax.numpy import asarray as asarray
from jax.numpy import empty_like as empty_like
from jax.numpy import full as full
from jax.numpy import full_like as full_like
from jax.numpy import linspace as linspace
from jax.numpy import meshgrid as meshgrid
from jax.numpy import ones_like as ones_like
from jax.numpy import tril as tril
from jax.numpy import triu as triu
from jax.numpy import zeros_like as zeros_like

# _higher_order.py
# isort: split
from jax.numpy import vectorize as vectorize



# ruff:noqa: F822

"""Quaxed :mod:`jax.scipy.linalg`.

This module wraps the functions in :external:`jax.scipy.linalg` with
:external:`quax.quaxify`. The wrapping happens dynamically through a
module-level ``__dir__`` and ``__getattr__``. The list of available functions is
in ``__all__`` and documented in the built-in :external:`jax.scipy.linalg`
library.

"""

__all__ = [
    "svd",
]

import sys
from collections.abc import Callable
from typing import Any

import jax.scipy.linalg
from quax import quaxify


def __dir__() -> list[str]:
    return sorted(__all__)


# TODO: better return type annotation
def __getattr__(name: str) -> Callable[..., Any]:
    if name not in __all__:
        msg = f"Cannot get {name} from quaxed.scipy.linalg."
        raise AttributeError(msg)

    # Quaxify the func
    func = quaxify(getattr(jax.scipy.linalg, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, func)

    return func
